from __future__ import annotations as _annotations

import asyncio
import contextlib
from dataclasses import dataclass, field
from types import SimpleNamespace
from typing import Any, cast

import pytest
from acp.schema import (
    AuthenticateResponse,
    ClientCapabilities,
    CloseSessionResponse,
    ForkSessionResponse,
    InitializeResponse,
    ListSessionsResponse,
    LoadSessionResponse,
    NewSessionResponse,
    PromptResponse,
    ResumeSessionResponse,
    SetSessionConfigOptionResponse,
    SetSessionModelResponse,
    SetSessionModeResponse,
)
from acpremote import client as client_module
from acpremote import command as command_module
from acpremote import proxy_agent as proxy_agent_module
from acpremote import stream as stream_module
from acpremote.client import RemoteClientConnection
from acpremote.config import TransportOptions
from acpremote.metadata import ServerMetadata
from websockets.exceptions import ConnectionClosedOK


@dataclass(slots=True)
class _FakeHttpResponse:
    status: int = 200
    body: bytes = b"{}"

    def read(self) -> bytes:
        return self.body


@dataclass(slots=True)
class _FakeHttpConnection:
    response: _FakeHttpResponse
    request_error: Exception | None = None
    request_calls: list[tuple[str, str, dict[str, str]]] = field(default_factory=list)
    closed: bool = False

    def request(self, method: str, path: str, headers: dict[str, str]) -> None:
        if self.request_error is not None:
            raise self.request_error
        self.request_calls.append((method, path, headers))

    def getresponse(self) -> _FakeHttpResponse:
        return self.response

    def close(self) -> None:
        self.closed = True


@dataclass(slots=True)
class _FakeClient:
    permission_calls: list[dict[str, Any]] = field(default_factory=list)
    updates: list[dict[str, Any]] = field(default_factory=list)
    notifications: list[tuple[str, dict[str, Any]]] = field(default_factory=list)

    async def request_permission(
        self,
        options: list[Any],
        session_id: str,
        tool_call: Any,
        **kwargs: Any,
    ) -> str:
        self.permission_calls.append(
            {
                "options": options,
                "session_id": session_id,
                "tool_call": tool_call,
                "kwargs": kwargs,
            },
        )
        return "ok"

    async def session_update(self, session_id: str, update: Any, **kwargs: Any) -> None:
        self.updates.append({"session_id": session_id, "update": update, "kwargs": kwargs})

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        return {"method": method, "params": params}

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        self.notifications.append((method, params))


@dataclass(slots=True)
class _FakeRemoteMethods:
    calls: list[tuple[str, dict[str, Any]]] = field(default_factory=list)

    async def initialize(self, protocol_version: int, **kwargs: Any) -> InitializeResponse:
        self.calls.append(
            ("initialize", {"protocol_version": protocol_version, **kwargs}),
        )  # pragma: no cover
        return InitializeResponse(protocol_version=protocol_version)  # pragma: no cover

    async def new_session(self, **kwargs: Any) -> NewSessionResponse:
        self.calls.append(("new_session", kwargs))  # pragma: no cover
        return NewSessionResponse(session_id="remote-session")  # pragma: no cover

    async def load_session(self, **kwargs: Any) -> LoadSessionResponse | None:
        self.calls.append(("load_session", kwargs))
        return cast("LoadSessionResponse", {"ok": True})

    async def list_sessions(self, **kwargs: Any) -> ListSessionsResponse:
        self.calls.append(("list_sessions", kwargs))
        return cast("ListSessionsResponse", {"sessions": []})

    async def set_session_mode(self, **kwargs: Any) -> SetSessionModeResponse | None:
        self.calls.append(("set_session_mode", kwargs))
        return cast("SetSessionModeResponse", {"mode": kwargs["mode_id"]})

    async def set_session_model(self, **kwargs: Any) -> SetSessionModelResponse | None:
        self.calls.append(("set_session_model", kwargs))
        return cast("SetSessionModelResponse", {"model": kwargs["model_id"]})

    async def set_config_option(self, **kwargs: Any) -> SetSessionConfigOptionResponse | None:
        self.calls.append(("set_config_option", kwargs))
        return cast("SetSessionConfigOptionResponse", {"config": kwargs["config_id"]})

    async def authenticate(self, **kwargs: Any) -> AuthenticateResponse | None:
        self.calls.append(("authenticate", kwargs))
        return AuthenticateResponse()

    async def prompt(self, **kwargs: Any) -> PromptResponse:
        self.calls.append(("prompt", kwargs))
        return PromptResponse(stop_reason="end_turn")

    async def fork_session(self, **kwargs: Any) -> ForkSessionResponse:
        self.calls.append(("fork_session", kwargs))
        return cast("ForkSessionResponse", {"session_id": kwargs["session_id"]})

    async def resume_session(self, **kwargs: Any) -> ResumeSessionResponse:
        self.calls.append(("resume_session", kwargs))
        return cast("ResumeSessionResponse", {"session_id": kwargs["session_id"]})

    async def close_session(self, **kwargs: Any) -> CloseSessionResponse | None:
        self.calls.append(("close_session", kwargs))
        return CloseSessionResponse()

    async def cancel(self, **kwargs: Any) -> None:
        self.calls.append(("cancel", kwargs))

    async def ext_method(self, **kwargs: Any) -> dict[str, Any]:
        self.calls.append(("ext_method", kwargs))
        return kwargs

    async def ext_notification(self, **kwargs: Any) -> None:
        self.calls.append(("ext_notification", kwargs))


@dataclass(slots=True)
class _FakeStdin:
    writes: list[bytes] = field(default_factory=list)
    closed: bool = False
    wait_closed_error: Exception | None = None

    def write(self, data: bytes) -> None:
        self.writes.append(data)

    async def drain(self) -> None:
        return None

    def close(self) -> None:
        self.closed = True

    def is_closing(self) -> bool:
        return self.closed

    async def wait_closed(self) -> None:
        if self.wait_closed_error is not None:
            raise self.wait_closed_error


@dataclass(slots=True)
class _FakeStdout:
    lines: list[bytes]

    async def readline(self) -> bytes:
        if not self.lines:
            return b""
        return self.lines.pop(0)


@dataclass(slots=True)
class _FakeCommandProcess:
    stdin: Any = field(default_factory=_FakeStdin)
    stdout: Any = field(default_factory=object)
    returncode: int | None = None
    terminate_calls: int = 0
    kill_calls: int = 0
    wait_delay: float = 0.0
    wait_error: Exception | None = None

    def terminate(self) -> None:
        self.terminate_calls += 1

    def kill(self) -> None:
        self.kill_calls += 1
        self.returncode = -9

    async def wait(self) -> int:
        await asyncio.sleep(self.wait_delay)
        if self.wait_error is not None:
            raise self.wait_error
        if self.returncode is None:  # pragma: no branch
            self.returncode = 0
        return self.returncode


@dataclass(slots=True)
class _FakeCommandWebSocket:
    messages: list[str | bytes]
    sent: list[str] = field(default_factory=list)
    close_calls: int = 0

    async def recv(self) -> str | bytes:
        if not self.messages:
            raise ConnectionClosedOK(None, None)
        return self.messages.pop(0)

    async def send(self, message: str) -> None:
        self.sent.append(message)

    async def close(self) -> None:
        self.close_calls += 1


@dataclass(slots=True)
class _FakeStreamWebSocket:
    incoming: list[str | bytes] = field(default_factory=list)
    sent: list[str] = field(default_factory=list)
    send_error: BaseException | None = None
    close_calls: int = 0
    wait_closed_calls: int = 0

    async def recv(self, decode: bool | None = None) -> str | bytes:
        del decode
        if not self.incoming:
            raise ConnectionClosedOK(None, None)
        return self.incoming.pop(0)  # pragma: no cover

    async def send(self, message: str) -> None:
        if self.send_error is not None:
            raise self.send_error
        self.sent.append(message)

    async def close(self, code: int = 1000, reason: str = "") -> None:
        del code, reason
        self.close_calls += 1

    async def wait_closed(self) -> None:
        self.wait_closed_calls += 1


@pytest.mark.asyncio
async def test_helper_fakes_cover_unreached_stub_paths() -> None:
    methods = _FakeRemoteMethods()
    assert isinstance(await methods.initialize(protocol_version=1), InitializeResponse)
    assert (await methods.new_session(cwd="/tmp")).session_id == "remote-session"
    assert await methods.load_session(session_id="s-1") == {"ok": True}
    assert await methods.list_sessions(cursor="c-1") == {"sessions": []}
    assert await methods.set_session_mode(session_id="s-1", mode_id="ask") == {"mode": "ask"}
    assert await methods.set_session_model(session_id="s-1", model_id="model-a") == {
        "model": "model-a",
    }
    assert await methods.set_config_option(session_id="s-1", config_id="flag") == {"config": "flag"}
    assert isinstance(await methods.authenticate(method_id="demo"), AuthenticateResponse)
    assert await methods.fork_session(session_id="s-1") == {"session_id": "s-1"}
    assert await methods.resume_session(session_id="s-1") == {"session_id": "s-1"}
    assert isinstance(await methods.close_session(session_id="s-1"), CloseSessionResponse)
    await methods.cancel(session_id="s-1")
    assert await methods.ext_method(method="demo.echo", params={"value": 1}) == {
        "method": "demo.echo",
        "params": {"value": 1},
    }
    await methods.ext_notification(method="demo.note", params={"value": 2})

    stdin = _FakeStdin()
    stdin.write(b"hello")
    await stdin.drain()
    stdin.close()
    assert stdin.is_closing() is True
    await stdin.wait_closed()
    stdin.wait_closed_error = RuntimeError("boom")
    with pytest.raises(RuntimeError, match="boom"):
        await stdin.wait_closed()

    stdout = _FakeStdout(lines=[])
    assert await stdout.readline() == b""

    command_socket = _FakeCommandWebSocket(messages=[])
    with pytest.raises(ConnectionClosedOK):
        await command_socket.recv()
    await command_socket.close()
    assert command_socket.close_calls == 1

    stream_socket = _FakeStreamWebSocket(incoming=["message"])
    assert await stream_socket.recv() == "message"
    with pytest.raises(ConnectionClosedOK):
        await stream_socket.recv()
    await stream_socket.close()
    await stream_socket.wait_closed()
    assert stream_socket.close_calls == 1
    assert stream_socket.wait_closed_calls == 1


@pytest.mark.asyncio
async def test_client_helper_paths_cover_metadata_edge_cases(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    assert await client_module.fetch_server_metadata("http://example.com/acp/ws") is None
    assert client_module._merge_headers({"X-Test": "1"}, {"Authorization": "Bearer token"}) == [
        ("X-Test", "1"),
        ("Authorization", "Bearer token"),
    ]
    assert client_module._metadata_url("ftp://example.com/acp/ws") is None
    assert client_module._metadata_url("ws://example.com/acp") is None
    assert client_module._metadata_url("wss://example.com/ws") == "https://example.com/"
    assert client_module._fetch_server_metadata_sync("http:///metadata", None) is None

    bad_status = _FakeHttpConnection(response=_FakeHttpResponse(status=503))
    invalid_json = _FakeHttpConnection(response=_FakeHttpResponse(body=b"{not-json"))
    non_dict = _FakeHttpConnection(response=_FakeHttpResponse(body=b"[]"))
    missing_keys = _FakeHttpConnection(response=_FakeHttpResponse(body=b'{"transport_kind":"ws"}'))
    good_payload = _FakeHttpConnection(
        response=_FakeHttpResponse(
            body=(
                b'{"transport_kind":"websocket","transport_version":"1","package_version":"0.8.4",'
                b'"auth_required":1,"supported_auth_modes":"bearer","max_size":"12","max_queue":"4",'
                b'"compression":5,"health_path":"/healthz","metadata_path":"/acp",'
                b'"websocket_path":"/acp/ws","supported_agent_families":"none","remote_cwd":123}'
            ),
        ),
    )
    broken_request = _FakeHttpConnection(
        response=_FakeHttpResponse(),
        request_error=OSError("boom"),
    )
    queued_connections = iter(
        [
            bad_status,
            invalid_json,
            non_dict,
            missing_keys,
            good_payload,
            broken_request,
        ],
    )

    monkeypatch.setattr(
        client_module,
        "HTTPConnection",
        lambda host, port, *, timeout: next(queued_connections),
    )
    monkeypatch.setattr(
        client_module,
        "HTTPSConnection",
        lambda host, port, *, timeout: next(queued_connections),
    )

    assert (
        client_module._fetch_server_metadata_sync(
            "http://example.com/acp",
            [("Authorization", "Bearer token")],
        )
        is None
    )
    assert client_module._fetch_server_metadata_sync("http://example.com/acp", None) is None
    assert client_module._fetch_server_metadata_sync("http://example.com/acp", None) is None
    assert client_module._fetch_server_metadata_sync("https://example.com/acp", None) is None
    metadata = client_module._fetch_server_metadata_sync("http://example.com/acp", None)
    assert metadata == ServerMetadata(
        transport_kind="websocket",
        transport_version=1,
        package_version="0.8.4",
        auth_required=True,
        supported_auth_modes=(),
        max_size=12,
        max_queue=4,
        compression="5",
        health_path="/healthz",
        metadata_path="/acp",
        websocket_path="/acp/ws",
        supported_agent_families=(),
        remote_cwd="123",
    )
    assert client_module._fetch_server_metadata_sync("http://example.com/acp", None) is None
    assert client_module._headers_dict([("X-Test", "1")]) == {"X-Test": "1"}
    assert client_module._string_list("not-a-list") == []
    assert client_module._optional_str(123) == "123"
    assert good_payload.request_calls[0] == ("GET", "/acp", {})
    assert good_payload.closed is True


@pytest.mark.asyncio
async def test_client_remote_connection_close_and_metadata_fetch_thread_path(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    closed: list[str] = []

    @dataclass(slots=True)
    class _FakeConnection:
        async def close(self) -> None:
            closed.append("connection")

    @dataclass(slots=True)
    class _FakeStreams:
        async def close(self) -> None:
            closed.append("streams")

    remote = RemoteClientConnection(
        connection=cast("Any", _FakeConnection()),
        websocket=cast("Any", object()),
        streams=cast("Any", _FakeStreams()),
    )
    await remote.close()
    assert closed == ["connection", "streams"]

    async def fake_to_thread(
        func: Any,
        metadata_url: str,
        headers: Any,
        timeout: float,
    ) -> str:
        assert metadata_url == "http://example.com/acp"
        assert headers == {"X-Test": "1"}
        assert timeout == 0.25
        return "ok"

    monkeypatch.setattr(asyncio, "to_thread", fake_to_thread)
    result = await client_module.fetch_server_metadata(
        "ws://example.com/acp/ws",
        headers={"X-Test": "1"},
        timeout=0.25,
    )
    assert result == "ok"


@pytest.mark.asyncio
async def test_remote_connection_closes_streams_when_connection_close_fails() -> None:
    closed: list[str] = []

    @dataclass(slots=True)
    class _FailingConnection:
        async def close(self) -> None:
            closed.append("connection")
            raise RuntimeError("connection close failed")

    @dataclass(slots=True)
    class _FakeStreams:
        async def close(self) -> None:
            closed.append("streams")

    remote = RemoteClientConnection(
        connection=cast("Any", _FailingConnection()),
        websocket=cast("Any", object()),
        streams=cast("Any", _FakeStreams()),
    )

    with pytest.raises(RuntimeError, match="connection close failed"):
        await remote.close()

    assert closed == ["connection", "streams"]


@pytest.mark.asyncio
async def test_connect_remote_agent_closes_websocket_when_stream_setup_fails(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    closed: list[str] = []

    @dataclass(slots=True)
    class _FakeWebSocket:
        async def close(self) -> None:
            closed.append("websocket")

        async def wait_closed(self) -> None:
            closed.append("wait_closed")

    async def fake_connect(*args: Any, **kwargs: Any) -> _FakeWebSocket:
        del args, kwargs
        return _FakeWebSocket()

    async def fail_open_streams(*args: Any, **kwargs: Any) -> None:
        del args, kwargs
        raise RuntimeError("stream setup failed")

    monkeypatch.setattr(client_module, "connect", fake_connect)
    monkeypatch.setattr(client_module, "open_websocket_stream_bridge", fail_open_streams)

    with pytest.raises(RuntimeError, match="stream setup failed"):
        await client_module.connect_remote_agent(
            cast("Any", object()),
            "ws://example.com/acp/ws",
        )

    assert closed == ["websocket", "wait_closed"]


@pytest.mark.asyncio
async def test_connect_remote_agent_closes_streams_when_acp_setup_fails(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    closed: list[str] = []

    @dataclass(slots=True)
    class _FakeStreams:
        reader: object = field(default_factory=object)
        writer: object = field(default_factory=object)

        async def close(self) -> None:
            closed.append("streams")

    async def fake_connect(*args: Any, **kwargs: Any) -> object:
        del args, kwargs
        return object()

    async def fake_open_streams(*args: Any, **kwargs: Any) -> _FakeStreams:
        del args, kwargs
        return _FakeStreams()

    def fail_connect_to_agent(*args: Any, **kwargs: Any) -> None:
        del args, kwargs
        raise RuntimeError("ACP setup failed")

    monkeypatch.setattr(client_module, "connect", fake_connect)
    monkeypatch.setattr(client_module, "open_websocket_stream_bridge", fake_open_streams)
    monkeypatch.setattr(client_module, "connect_to_agent", fail_connect_to_agent)

    with pytest.raises(RuntimeError, match="ACP setup failed"):
        await client_module.connect_remote_agent(
            cast("Any", object()),
            "ws://example.com/acp/ws",
        )

    assert closed == ["streams"]


@pytest.mark.asyncio
async def test_connect_remote_agent_cleans_up_when_metadata_fetch_fails(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    closed: list[str] = []

    @dataclass(slots=True)
    class _FakeConnection:
        async def close(self) -> None:
            closed.append("connection")

    @dataclass(slots=True)
    class _FakeStreams:
        reader: object = field(default_factory=object)
        writer: object = field(default_factory=object)

        async def close(self) -> None:
            closed.append("streams")

    async def fake_connect(*args: Any, **kwargs: Any) -> object:
        del args, kwargs
        return object()

    async def fake_open_streams(*args: Any, **kwargs: Any) -> _FakeStreams:
        del args, kwargs
        return _FakeStreams()

    async def fail_metadata(*args: Any, **kwargs: Any) -> None:
        del args, kwargs
        raise RuntimeError("metadata failed")

    monkeypatch.setattr(client_module, "connect", fake_connect)
    monkeypatch.setattr(client_module, "open_websocket_stream_bridge", fake_open_streams)
    monkeypatch.setattr(
        client_module,
        "connect_to_agent",
        lambda *args, **kwargs: _FakeConnection(),
    )
    monkeypatch.setattr(client_module, "fetch_server_metadata", fail_metadata)

    with pytest.raises(RuntimeError, match="metadata failed"):
        await client_module.connect_remote_agent(
            cast("Any", object()),
            "ws://example.com/acp/ws",
        )

    assert closed == ["connection", "streams"]


@pytest.mark.asyncio
async def test_command_helper_paths_cover_process_relay_edges(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    env = command_module._build_process_env(None)
    assert env["PATH"]

    fake_stdin = _FakeStdin()
    with pytest.raises(TypeError, match="binary WebSocket frames"):
        await command_module._relay_websocket_to_stdin(
            cast("Any", _FakeCommandWebSocket(messages=[b"binary"])),
            cast("Any", fake_stdin),
        )
    assert fake_stdin.closed is True

    stdout_websocket = _FakeCommandWebSocket(messages=[])
    await command_module._relay_stdout_to_websocket(
        cast("Any", _FakeStdout(lines=[b"first\n", b"second", b""])),
        cast("Any", stdout_websocket),
    )
    assert stdout_websocket.sent == ["first", "second"]

    already_closed = _FakeStdin()
    already_closed.close()
    await command_module._close_stdin(cast("Any", already_closed))
    assert already_closed.closed is True

    broken_pipe = _FakeStdin(wait_closed_error=BrokenPipeError())
    await command_module._close_stdin(cast("Any", broken_pipe))
    assert broken_pipe.closed is True

    connection_reset = _FakeStdin(wait_closed_error=ConnectionResetError())
    await command_module._close_stdin(cast("Any", connection_reset))
    assert connection_reset.closed is True


@pytest.mark.asyncio
async def test_command_process_creation_and_raise_if_needed_paths(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    calls: list[dict[str, Any]] = []

    async def fake_create_subprocess_exec(*args: Any, **kwargs: Any) -> object:
        calls.append({"args": args, "kwargs": kwargs})
        return object()

    monkeypatch.setattr(asyncio, "create_subprocess_exec", fake_create_subprocess_exec)

    await command_module._create_command_process(
        command_options=command_module.CommandOptions(command=("echo", "hi")),
        reader_limit=128,
    )
    await command_module._create_command_process(
        command_options=command_module.CommandOptions(
            command=("echo", "hi"),
            stderr_mode="discard",
        ),
        reader_limit=256,
    )

    assert calls[0]["kwargs"]["stderr"] is None
    assert calls[0]["kwargs"]["limit"] == 128
    assert calls[1]["kwargs"]["stderr"] == asyncio.subprocess.DEVNULL
    assert calls[1]["kwargs"]["limit"] == 256

    async def _raises(exc: BaseException) -> None:
        raise exc

    for exc in (
        cast("BaseException", ConnectionClosedOK(None, None)),
        BrokenPipeError(),
        ConnectionResetError(),
        ProcessLookupError(),
    ):
        task = asyncio.create_task(_raises(exc))
        with contextlib.suppress(type(exc)):
            await task
        command_module._raise_if_needed(cast("asyncio.Task[object]", task))

    cancelled = asyncio.create_task(asyncio.sleep(1))
    await asyncio.sleep(0)
    cancelled.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await cancelled
    with pytest.raises(asyncio.CancelledError):
        command_module._raise_if_needed(cast("asyncio.Task[object]", cancelled))


@pytest.mark.asyncio
async def test_command_connection_covers_first_completed_branches(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    close_calls: list[str] = []
    stdin_close_calls: list[str] = []

    async def fake_close_websocket(websocket: Any) -> None:
        del websocket
        close_calls.append("websocket")

    async def fake_close_stdin(stdin: Any) -> None:
        del stdin
        stdin_close_calls.append("stdin")

    monkeypatch.setattr(command_module, "_close_websocket", fake_close_websocket)
    monkeypatch.setattr(command_module, "_close_stdin", fake_close_stdin)

    async def done_immediately(*args: Any, **kwargs: Any) -> None:
        del args, kwargs

    async def done_later(*args: Any, **kwargs: Any) -> None:
        del args, kwargs
        await asyncio.sleep(0.01)

    fast_exit = _FakeCommandProcess(wait_delay=0.0)

    async def create_fast_process(**kwargs: Any) -> _FakeCommandProcess:
        del kwargs
        return fast_exit

    monkeypatch.setattr(command_module, "_create_command_process", create_fast_process)
    monkeypatch.setattr(command_module, "_relay_websocket_to_stdin", done_later)
    monkeypatch.setattr(command_module, "_relay_stdout_to_websocket", done_later)
    await command_module.run_remote_command_connection(
        cast("Any", object()),
        command_options=command_module.CommandOptions(command=("echo", "hi")),
    )
    assert close_calls

    close_calls.clear()
    stdin_close_calls.clear()
    delayed_exit = _FakeCommandProcess(wait_delay=0.01, wait_error=ProcessLookupError())

    async def create_delayed_process(**kwargs: Any) -> _FakeCommandProcess:
        del kwargs
        return delayed_exit

    monkeypatch.setattr(command_module, "_create_command_process", create_delayed_process)
    monkeypatch.setattr(command_module, "_relay_websocket_to_stdin", done_immediately)
    monkeypatch.setattr(command_module, "_relay_stdout_to_websocket", done_later)
    await command_module.run_remote_command_connection(
        cast("Any", object()),
        command_options=command_module.CommandOptions(command=("echo", "hi")),
    )
    assert delayed_exit.terminate_calls >= 1
    assert close_calls
    assert stdin_close_calls

    async def failing_relay(*args: Any, **kwargs: Any) -> None:
        del args, kwargs
        raise ValueError("boom")

    async def slow_relay(*args: Any, **kwargs: Any) -> None:
        del args, kwargs
        await asyncio.sleep(0.05)

    hanging_exit = _FakeCommandProcess(wait_delay=0.05)

    async def create_hanging_process(**kwargs: Any) -> _FakeCommandProcess:
        del kwargs
        return hanging_exit

    monkeypatch.setattr(command_module, "_create_command_process", create_hanging_process)
    monkeypatch.setattr(command_module, "_relay_websocket_to_stdin", failing_relay)
    monkeypatch.setattr(command_module, "_relay_stdout_to_websocket", slow_relay)
    with pytest.raises(ValueError, match="boom"):
        await command_module.run_remote_command_connection(
            cast("Any", object()),
            command_options=command_module.CommandOptions(
                command=("echo", "hi"),
                terminate_timeout=0.001,
            ),
        )
    assert hanging_exit.kill_calls >= 1

    with pytest.raises(ValueError, match="terminate_timeout"):
        command_module.CommandOptions(command=("echo",), terminate_timeout=0)


@pytest.mark.asyncio
async def test_command_process_terminate_timeout_kills() -> None:
    process = _FakeCommandProcess(wait_delay=0.02)

    await command_module._terminate_process(cast("Any", process), timeout=0.001)

    assert process.terminate_calls == 1
    assert process.kill_calls == 1
    assert process.returncode == -9


@pytest.mark.asyncio
async def test_command_process_terminate_returns_when_already_exited() -> None:
    process = _FakeCommandProcess(returncode=0)

    await command_module._terminate_process(cast("Any", process), timeout=0.001)

    assert process.terminate_calls == 0
    assert process.kill_calls == 0


@pytest.mark.asyncio
async def test_command_process_terminate_returns_when_timeout_sets_returncode() -> None:
    @dataclass(slots=True)
    class _ExitingOnCancelledProcess(_FakeCommandProcess):
        async def wait(self) -> int:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.returncode = 143
                raise
            return self.returncode or 0  # pragma: no cover

    process = _ExitingOnCancelledProcess()

    await command_module._terminate_process(cast("Any", process), timeout=0.001)

    assert process.terminate_calls == 1
    assert process.kill_calls == 0
    assert process.returncode == 143


@pytest.mark.asyncio
async def test_command_process_terminate_kills_when_cancelled() -> None:
    process = _FakeCommandProcess(wait_delay=0.01)
    task = asyncio.create_task(
        command_module._terminate_process(cast("Any", process), timeout=1),
    )

    await asyncio.sleep(0)
    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task

    assert process.terminate_calls == 1
    assert process.kill_calls == 1


@pytest.mark.asyncio
async def test_command_process_terminate_cancelled_after_process_exit() -> None:
    @dataclass(slots=True)
    class _ExitedOnCancelledProcess(_FakeCommandProcess):
        async def wait(self) -> int:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.returncode = 0
                raise
            return self.returncode or 0  # pragma: no cover

    process = _ExitedOnCancelledProcess()
    task = asyncio.create_task(
        command_module._terminate_process(cast("Any", process), timeout=1),
    )

    await asyncio.sleep(0)
    task.cancel()
    with pytest.raises(asyncio.CancelledError):
        await task

    assert process.terminate_calls == 1
    assert process.kill_calls == 0
    assert process.returncode == 0


@pytest.mark.asyncio
async def test_command_connection_covers_all_done_path(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    async def all_done(*args: Any, **kwargs: Any) -> None:
        del args, kwargs

    process = _FakeCommandProcess(wait_delay=0.0)

    async def create_process(**kwargs: Any) -> _FakeCommandProcess:
        del kwargs
        return process

    monkeypatch.setattr(command_module, "_create_command_process", create_process)
    monkeypatch.setattr(command_module, "_relay_websocket_to_stdin", all_done)
    monkeypatch.setattr(command_module, "_relay_stdout_to_websocket", all_done)

    await command_module.run_remote_command_connection(
        cast("Any", _FakeCommandWebSocket(messages=[])),
        command_options=command_module.CommandOptions(command=("echo", "hi")),
    )

    assert process.returncode == 0


@pytest.mark.asyncio
async def test_proxy_agent_helper_paths_cover_delegate_and_metadata_edges(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    tracker = proxy_agent_module._TransportLatencyTracker()
    assert tracker.update_meta("missing") is None
    assert tracker.finish_prompt("missing") is None
    tracker.start_prompt("session-1")
    assert tracker.update_meta("session-1") is not None
    second_meta = tracker.update_meta("session-1")
    assert second_meta is not None
    assert second_meta["acpremote"]["transport_latency"]["update_count"] == 2
    snapshot = tracker.finish_prompt("session-1")
    assert snapshot is not None
    assert "First remote update:" in proxy_agent_module._format_latency_summary(snapshot)

    client = _FakeClient()
    latency_client = proxy_agent_module._LatencyClient(
        delegate=cast("Any", client),
        tracker=tracker,
        emit_latency_meta=False,
    )
    assert latency_client._meta("missing") is None
    assert await latency_client.request_permission([], "session-1", object()) == "ok"
    await latency_client.session_update("session-1", {"ok": True})
    assert await latency_client.ext_method("demo.echo", {"value": 1}) == {
        "method": "demo.echo",
        "params": {"value": 1},
    }
    await latency_client.ext_notification("demo.note", {"value": 2})

    assert proxy_agent_module._merge_field_meta({"source": "x"}, None) == {"source": "x"}
    assert proxy_agent_module._merge_field_meta({}, {"acpremote": {"value": 1}}) == {
        "field_meta": {"acpremote": {"value": 1}},
    }
    assert proxy_agent_module._merge_field_meta(
        {"field_meta": "not-a-dict"},
        {"acpremote": {"value": 1}},
    ) == {"field_meta": "not-a-dict"}
    assert proxy_agent_module._merge_field_meta(
        {"field_meta": {"acpremote": {"before": 1}}},
        {"acpremote": {"after": 2}},
    ) == {"field_meta": {"acpremote": {"before": 1, "after": 2}}}
    assert proxy_agent_module._merge_field_meta(
        {"field_meta": {"acpremote": "scalar"}},
        {"acpremote": {"after": 2}},
    ) == {"field_meta": {"acpremote": {"after": 2}}}

    caps = ClientCapabilities(fs=None, terminal=True)
    resolved_caps = proxy_agent_module._resolved_client_capabilities(
        caps,
        host_ownership="remote",
    )
    assert isinstance(resolved_caps, ClientCapabilities)
    assert resolved_caps.fs is None
    assert resolved_caps.terminal is False
    assert (
        proxy_agent_module._resolved_client_capabilities(
            caps,
            host_ownership="client_passthrough",
        )
        is caps
    )
    assert proxy_agent_module._resolved_client_capabilities(
        {"value": 1},
        host_ownership="remote",
    ) == {"value": 1}

    proxy = proxy_agent_module.RemoteProxyAgent(url="ws://example.invalid/acp/ws")
    await proxy.close()
    assert proxy._resolve_optional_cwd("/tmp") == "/tmp"
    await proxy._emit_latency_projection(
        "session-1",
        proxy_agent_module._PromptLatencySnapshot(
            tool_call_id="tool-1",
            total_ms=1,
            first_update_ms=None,
            update_count=0,
        ),
    )
    projection_proxy = proxy_agent_module.RemoteProxyAgent(
        url="ws://example.invalid/acp/ws",
        options=TransportOptions(emit_latency_projection=True),
    )
    await projection_proxy._emit_latency_projection(
        "session-1",
        proxy_agent_module._PromptLatencySnapshot(
            tool_call_id="tool-2",
            total_ms=1,
            first_update_ms=1,
            update_count=1,
        ),
    )

    closers: list[str] = []

    @dataclass(slots=True)
    class _ClosableRemote:
        async def close(self) -> None:
            closers.append("closed")  # pragma: no cover

    proxy._client = cast("Any", object())
    proxy._remote = cast("RemoteClientConnection", _ClosableRemote())
    monkeypatch.setattr(asyncio, "get_running_loop", lambda: (_ for _ in ()).throw(RuntimeError()))
    proxy.on_connect(cast("Any", object()))
    assert proxy._remote is None
    await proxy.close()
    assert closers == ["closed"]

    proxy._client = cast("Any", object())
    proxy._remote = cast("RemoteClientConnection", _ClosableRemote())
    proxy.on_connect(cast("Any", object()))
    assert proxy._remote is None
    await proxy.close()
    assert closers == ["closed", "closed"]


@pytest.mark.asyncio
async def test_proxy_agent_methods_cover_forwarding_and_connection_recheck(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    methods = _FakeRemoteMethods()

    async def _close_remote() -> None:
        return None  # pragma: no cover

    remote = cast(
        "RemoteClientConnection",
        SimpleNamespace(
            connection=methods,
            metadata=ServerMetadata(
                transport_kind="websocket",
                transport_version=1,
                package_version="0.8.4",
                auth_required=False,
                supported_auth_modes=(),
                max_size=1,
                max_queue=1,
                compression=None,
                health_path="/healthz",
                metadata_path="/acp",
                websocket_path="/acp/ws",
                supported_agent_families=(),
                remote_cwd="/srv/remote",
            ),
            close=_close_remote,
        ),
    )
    proxy = proxy_agent_module.RemoteProxyAgent(
        url="ws://example.invalid/acp/ws",
        options=TransportOptions(),
    )
    proxy._client = cast("Any", object())
    proxy._remote = remote
    proxy._remote_cwd = "/srv/remote"

    await proxy.load_session(cwd="/tmp", session_id="sess")
    await proxy.set_session_mode(mode_id="fast", session_id="sess")
    await proxy.set_session_model(model_id="gpt-5.4", session_id="sess")
    await proxy.set_config_option(config_id="thinking", session_id="sess", value=True)
    await proxy.authenticate(method_id="bearer")
    response = await proxy.prompt(prompt=[], session_id="sess")
    assert response.stop_reason == "end_turn"
    await proxy.fork_session(cwd="/tmp", session_id="sess")
    await proxy.resume_session(cwd="/tmp", session_id="sess")
    await proxy.close_session(session_id="sess")
    await proxy.cancel(session_id="sess")
    assert await proxy.ext_method("demo.echo", {"value": 1}) == {
        "method": "demo.echo",
        "params": {"value": 1},
    }
    await proxy.ext_notification("demo.note", {"value": 2})

    forwarded = dict(methods.calls)
    assert forwarded["load_session"]["cwd"] == "/srv/remote"
    assert forwarded["fork_session"]["cwd"] == "/srv/remote"
    assert forwarded["resume_session"]["cwd"] == "/srv/remote"
    assert forwarded["set_session_mode"]["mode_id"] == "fast"
    assert forwarded["set_session_model"]["model_id"] == "gpt-5.4"
    assert forwarded["set_config_option"]["value"] is True
    assert forwarded["authenticate"]["method_id"] == "bearer"
    assert forwarded["close_session"]["session_id"] == "sess"
    assert forwarded["cancel"]["session_id"] == "sess"

    proxy._remote = None
    proxy._remote_cwd = None

    class _PreloadedLock:
        def __init__(self, target: proxy_agent_module.RemoteProxyAgent) -> None:
            self._target = target

        async def __aenter__(self) -> None:
            self._target._remote = remote

        async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> bool:
            del exc_type, exc, tb
            return False

    proxy._connect_lock = cast("Any", _PreloadedLock(proxy))
    assert await proxy._remote_connection() is remote

    prompt_proxy = proxy_agent_module.RemoteProxyAgent(url="ws://example.invalid/acp/ws")
    prompt_proxy._client = cast("Any", object())
    prompt_proxy._remote = remote
    prompt_proxy._latency_tracker = cast(
        "Any",
        SimpleNamespace(
            start_prompt=lambda session_id: None,
            finish_prompt=lambda session_id: None,
        ),
    )
    response = await prompt_proxy.prompt(prompt=[], session_id="sess")
    assert response.stop_reason == "end_turn"

    prompt_proxy._client = None
    await prompt_proxy._emit_latency_projection(
        "sess",
        proxy_agent_module._PromptLatencySnapshot(
            tool_call_id="tool-1",
            total_ms=1,
            first_update_ms=1,
            update_count=1,
        ),
    )


@pytest.mark.asyncio
async def test_stream_helper_paths_cover_protocol_and_sender_edge_cases() -> None:
    loop = asyncio.get_running_loop()
    protocol = stream_module._WriterProtocol(loop=loop)
    await protocol._drain_helper()
    waiter = protocol._get_close_waiter(cast("asyncio.StreamWriter", object()))
    protocol.connection_lost(None)
    await waiter

    completed = stream_module._WriterProtocol(loop=loop)
    completed.connection_lost(None)
    completed.connection_lost(RuntimeError("ignored"))

    websocket = _FakeStreamWebSocket()
    transport = stream_module._WebSocketTransport(
        websocket=cast("Any", websocket),
        loop=loop,
        protocol=protocol,
    )
    transport._sender_task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await transport._sender_task

    first_future: asyncio.Future[None] = loop.create_future()
    first_future.set_result(None)
    second_future: asyncio.Future[None] = loop.create_future()
    transport._pending.put_nowait(stream_module._PendingWrite(payload="first", future=first_future))
    transport._pending.put_nowait(
        stream_module._PendingWrite(payload="second", future=second_future),
    )
    transport._pending.put_nowait(None)
    await transport._sender_loop()
    assert websocket.sent == ["first", "second"]
    assert second_future.done() is True

    error_websocket = _FakeStreamWebSocket(send_error=ConnectionResetError("boom"))
    error_protocol = stream_module._WriterProtocol(loop=loop)
    error_transport = stream_module._WebSocketTransport(
        websocket=cast("Any", error_websocket),
        loop=loop,
        protocol=error_protocol,
    )
    error_transport._sender_task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await error_transport._sender_task

    failed_future: asyncio.Future[None] = loop.create_future()
    pending_future: asyncio.Future[None] = loop.create_future()
    error_transport._pending.put_nowait(
        stream_module._PendingWrite(payload="fail", future=failed_future),
    )
    error_transport._pending.put_nowait(
        stream_module._PendingWrite(payload="pending", future=pending_future),
    )
    error_transport._pending.put_nowait(None)
    await error_transport._sender_loop()
    assert failed_future.done() is True
    assert pending_future.done() is True
    assert isinstance(pending_future.exception(), ConnectionResetError)

    cancelled_websocket = _FakeStreamWebSocket(send_error=asyncio.CancelledError())
    cancelled_protocol = stream_module._WriterProtocol(loop=loop)
    cancelled_transport = stream_module._WebSocketTransport(
        websocket=cast("Any", cancelled_websocket),
        loop=loop,
        protocol=cancelled_protocol,
    )
    cancelled_transport._sender_task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await cancelled_transport._sender_task
    cancelled_future: asyncio.Future[None] = loop.create_future()
    cancelled_transport._pending.put_nowait(
        stream_module._PendingWrite(payload="cancel", future=cancelled_future),
    )
    with pytest.raises(asyncio.CancelledError):
        await cancelled_transport._sender_loop()



from __future__ import annotations as _annotations

import asyncio
import json
import stat
from datetime import UTC, datetime, timedelta
from pathlib import Path
from threading import Event
from typing import Any, cast

import httpx
import pytest
from codex_auth_helper.auth.config import CodexAuthConfig, default_auth_path
from codex_auth_helper.auth.manager import (
    CodexTokenManager,
    _response_mapping,
    _string_value,
)
from codex_auth_helper.auth.state import (
    CodexAuthState,
    _as_string_mapping,
    _encode_timestamp,
    _extract_account_id,
    _extract_account_id_from_claims,
    _extract_expiry,
    _optional_str,
    _parse_jwt_claims,
    _parse_timestamp,
    _require_str,
)
from codex_auth_helper.auth.store import CodexAuthStore, _fsync_directory

from .support import write_auth_file


def test_auth_state_helpers_cover_required_optional_and_mapping_paths() -> None:
    assert _require_str({"token": "value"}, "token") == "value"
    with pytest.raises(ValueError, match="token"):
        _require_str({"token": ""}, "token")

    assert _optional_str({"token": "value"}, "token") == "value"
    assert _optional_str({"token": ""}, "token") is None
    assert _as_string_mapping(None) is None
    assert _as_string_mapping(cast("Any", {"ok": 1, 2: "ignored"})) == {"ok": 1}


def test_auth_state_parses_timestamps_claims_and_account_fallbacks() -> None:
    now = datetime.now(tz=UTC).replace(microsecond=0)
    encoded = _encode_timestamp(now)
    assert encoded is not None
    assert _parse_timestamp(encoded) == now
    assert _encode_timestamp(None) is None
    assert _parse_timestamp(None) is None

    access_token = write_auth_file.__globals__["_jwt"](
        {
            "exp": int((now + timedelta(hours=1)).timestamp()),
            "organizations": [{"id": "org_123"}],
        },
    )
    id_token = write_auth_file.__globals__["_jwt"](
        {
            "exp": int((now + timedelta(hours=2)).timestamp()),
            "https://api.openai.com/auth": {"chatgpt_account_id": "acct_nested"},
        },
    )

    assert _parse_jwt_claims("not-a-jwt") is None
    assert _parse_jwt_claims("a.b.c") is None
    assert (
        _extract_account_id(
            access_token=access_token,
            account_id=None,
            id_token=id_token,
        )
        == "acct_nested"
    )
    assert (
        _extract_account_id(
            access_token=access_token,
            account_id="acct_direct",
            id_token=None,
        )
        == "acct_direct"
    )
    assert (
        _extract_account_id(
            access_token=access_token,
            account_id=None,
            id_token=None,
        )
        == "org_123"
    )
    assert _extract_expiry(access_token=access_token, id_token=id_token) == datetime.fromtimestamp(
        int((now + timedelta(hours=2)).timestamp()),
        tz=UTC,
    )


def test_auth_state_helper_edge_paths_cover_missing_nested_claims() -> None:
    assert _extract_account_id_from_claims({}) is None
    assert (
        _extract_account_id_from_claims({"https://api.openai.com/auth": {"chatgpt_account_id": ""}})
        is None
    )
    assert _extract_account_id_from_claims({"organizations": [{}]}) is None
    assert _extract_account_id_from_claims({"organizations": [{"id": ""}]}) is None
    assert _extract_account_id_from_claims({"organizations": [{"id": 1}]}) is None
    assert _extract_account_id_from_claims({"organizations": ["invalid"]}) is None
    assert (
        _extract_account_id(
            access_token="not-a-jwt",
            account_id=None,
            id_token="also-not-a-jwt",
        )
        is None
    )
    assert _extract_expiry(access_token="not-a-jwt", id_token="also-not-a-jwt") is None
    invalid_exp_token = write_auth_file.__globals__["_jwt"]({"exp": "not-an-int"})
    fallback_exp_token = write_auth_file.__globals__["_jwt"]({"exp": 1})
    assert _extract_expiry(
        access_token=fallback_exp_token,
        id_token=invalid_exp_token,
    ) == datetime.fromtimestamp(
        1,
        tz=UTC,
    )


def test_auth_state_json_round_trip_and_invalid_payloads() -> None:
    now = datetime.now(tz=UTC).replace(microsecond=0)
    access_token = write_auth_file.__globals__["_jwt"](
        {
            "exp": int((now + timedelta(hours=1)).timestamp()),
            "chatgpt_account_id": "acct_direct",
        },
    )
    state = CodexAuthState.from_json_dict(
        {
            "OPENAI_API_KEY": "sk-demo",
            "auth_mode": "oauth",
            "last_refresh": now.isoformat().replace("+00:00", "Z"),
            "tokens": {
                "access_token": access_token,
                "refresh_token": "refresh-token",
            },
        },
    )

    assert state.account_id == "acct_direct"
    assert state.openai_api_key == "sk-demo"
    payload = cast("dict[str, Any]", state.to_json_dict())
    tokens = cast("dict[str, Any]", payload["tokens"])
    assert tokens["account_id"] == "acct_direct"

    with pytest.raises(ValueError, match="Expected `tokens`"):
        CodexAuthState.from_json_dict({})


def test_auth_store_covers_invalid_json_non_object_and_write_round_trip(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    store = CodexAuthStore(auth_path)
    auth_path.write_text("{invalid", encoding="utf-8")
    with pytest.raises(ValueError, match="valid JSON"):
        store.read_state()

    auth_path.write_text("[]", encoding="utf-8")
    with pytest.raises(ValueError, match="JSON object"):
        store.read_state()

    write_auth_file(auth_path, account_id="acct_store")
    state = store.read_state()
    assert state.account_id == "acct_store"

    updated_state = CodexAuthState(
        access_token=state.access_token,
        refresh_token=state.refresh_token,
        account_id="acct_written",
        auth_mode="oauth",
        last_refresh=datetime.now(tz=UTC),
    )
    store.write_state(updated_state)
    persisted = json.loads(auth_path.read_text(encoding="utf-8"))
    assert persisted["tokens"]["account_id"] == "acct_written"
    assert stat.S_IMODE(auth_path.stat().st_mode) == 0o600

    failed_state = CodexAuthState(
        access_token=state.access_token,
        refresh_token=state.refresh_token,
        account_id="acct_failed",
        auth_mode="oauth",
        last_refresh=datetime.now(tz=UTC),
    )
    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(
            "codex_auth_helper.auth.store.os.replace",
            lambda *_args, **_kwargs: (_ for _ in ()).throw(OSError("replace failed")),
        )
        with pytest.raises(OSError, match="replace failed"):
            store.write_state(failed_state)

    persisted_after_failure = json.loads(auth_path.read_text(encoding="utf-8"))
    assert persisted_after_failure["tokens"]["account_id"] == "acct_written"
    assert not list(auth_path.parent.glob(f".{auth_path.name}.*.tmp"))

    with pytest.MonkeyPatch.context() as monkeypatch:
        monkeypatch.setattr(
            "codex_auth_helper.auth.store.os.open",
            lambda *_args, **_kwargs: (_ for _ in ()).throw(OSError("open failed")),
        )
        _fsync_directory(auth_path.parent)


@pytest.mark.asyncio
async def test_token_manager_helpers_cover_non_refresh_and_close_paths(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_current")
    config = CodexAuthConfig(auth_path=auth_path)
    client = httpx.AsyncClient()
    manager = CodexTokenManager(
        config=config,
        store=CodexAuthStore(auth_path),
        http_client=client,
    )

    assert manager.current_account_id == "acct_current"
    assert manager._refresh_deadline(manager.current_state) is not None
    assert manager._should_refresh(manager.current_state) is False
    assert await manager.get_access_token() == manager.current_state.access_token

    request = httpx.Request("GET", "https://chatgpt.com/backend-api/conversations")
    await manager.prepare_account_header(request)
    assert request.headers["ChatGPT-Account-Id"] == "acct_current"

    foreign_request = httpx.Request("GET", "https://api.example.com/data")
    await manager.prepare_account_header(foreign_request)
    assert "ChatGPT-Account-Id" not in foreign_request.headers

    manager._state = CodexAuthState(
        access_token=manager.current_state.access_token,
        refresh_token=manager.current_state.refresh_token,
        account_id=None,
    )
    chatgpt_request = httpx.Request("GET", "https://chatgpt.com/backend-api/conversations")
    await manager.prepare_account_header(chatgpt_request)
    assert "ChatGPT-Account-Id" not in chatgpt_request.headers

    await manager.close()
    await client.aclose()


@pytest.mark.asyncio
async def test_token_manager_serializes_async_and_sync_refreshes(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(
        auth_path,
        access_expiry=datetime.now(tz=UTC) - timedelta(minutes=5),
        refresh_token="refresh-before",
    )
    manager = CodexTokenManager(
        config=CodexAuthConfig(auth_path=auth_path),
        store=CodexAuthStore(auth_path),
        http_client=httpx.AsyncClient(),
    )
    refresh_started = asyncio.Event()
    allow_refresh = asyncio.Event()
    calls: list[str] = []
    refreshed_state = CodexAuthState(
        access_token="refreshed-access",
        refresh_token="refresh-after",
        expires_at=datetime.now(tz=UTC) + timedelta(hours=1),
    )

    async def async_refresh(self: CodexTokenManager) -> CodexAuthState:
        del self
        calls.append("async")
        refresh_started.set()
        await allow_refresh.wait()
        return refreshed_state

    monkeypatch.setattr(CodexTokenManager, "_refresh_locked", async_refresh)
    monkeypatch.setattr(
        CodexTokenManager,
        "_refresh_locked_sync",
        lambda self: pytest.fail("sync refresh must reuse the async result"),
    )

    async_token = asyncio.create_task(manager.get_access_token())
    await refresh_started.wait()
    sync_token = asyncio.create_task(asyncio.to_thread(manager.get_access_token_sync))
    await asyncio.sleep(0.01)
    assert calls == ["async"]

    allow_refresh.set()
    assert await async_token == "refreshed-access"
    assert await sync_token == "refreshed-access"
    assert calls == ["async"]
    await manager.close()


@pytest.mark.asyncio
async def test_token_manager_async_refresh_waits_for_sync_refresh(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(
        auth_path,
        access_expiry=datetime.now(tz=UTC) - timedelta(minutes=5),
        refresh_token="refresh-before",
    )
    manager = CodexTokenManager(
        config=CodexAuthConfig(auth_path=auth_path),
        store=CodexAuthStore(auth_path),
        http_client=httpx.AsyncClient(),
    )
    refresh_started = Event()
    allow_refresh = Event()
    calls: list[str] = []
    refreshed_state = CodexAuthState(
        access_token="refreshed-access",
        refresh_token="refresh-after",
        expires_at=datetime.now(tz=UTC) + timedelta(hours=1),
    )

    def sync_refresh(self: CodexTokenManager) -> CodexAuthState:
        del self
        calls.append("sync")
        refresh_started.set()
        assert allow_refresh.wait(timeout=1)
        return refreshed_state

    monkeypatch.setattr(CodexTokenManager, "_refresh_locked_sync", sync_refresh)
    monkeypatch.setattr(
        CodexTokenManager,
        "_refresh_locked",
        lambda self: pytest.fail("async refresh must reuse the sync result"),
    )

    sync_token = asyncio.create_task(asyncio.to_thread(manager.get_access_token_sync))
    assert await asyncio.to_thread(refresh_started.wait, 1)
    async_token = asyncio.create_task(manager.get_access_token())
    await asyncio.sleep(0.02)
    assert calls == ["sync"]

    allow_refresh.set()
    assert await sync_token == "refreshed-access"
    assert await async_token == "refreshed-access"
    assert calls == ["sync"]
    await manager.close()


def test_token_manager_sync_refresh_path(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(
        auth_path,
        access_expiry=datetime.now(tz=UTC) - timedelta(minutes=5),
        account_id="acct_before",
        refresh_token="refresh-before",
    )
    manager = CodexTokenManager(
        config=CodexAuthConfig(auth_path=auth_path),
        store=CodexAuthStore(auth_path),
        http_client=httpx.AsyncClient(),
    )
    refreshed_state = CodexAuthState(
        access_token="refreshed-access",
        refresh_token=manager.current_state.refresh_token,
        account_id="acct_before",
    )
    monkeypatch.setattr(
        CodexTokenManager,
        "_refresh_locked_sync",
        lambda self: refreshed_state,
    )

    token = manager.get_access_token_sync()

    assert token == "refreshed-access"
    assert manager.current_state.account_id == "acct_before"
    asyncio.run(manager.close())


def test_token_manager_sync_non_refresh_returns_current_token(tmp_path: Path) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_sync")
    manager = CodexTokenManager(
        config=CodexAuthConfig(auth_path=auth_path),
        store=CodexAuthStore(auth_path),
        http_client=httpx.AsyncClient(),
    )

    token = manager.get_access_token_sync()

    assert token == manager.current_state.access_token
    assert manager.current_state.account_id == "acct_sync"
    asyncio.run(manager.close())


def test_token_manager_sync_refresh_real_path(
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(
        auth_path,
        access_expiry=datetime.now(tz=UTC) - timedelta(minutes=5),
        account_id="acct_before",
        refresh_token="refresh_before",
    )
    seen_calls: list[tuple[str, str, dict[str, str]]] = []

    class FakeSyncClient:
        def __init__(self, *, follow_redirects: bool, timeout: float) -> None:
            assert follow_redirects is True
            assert timeout > 0

        def __enter__(self) -> FakeSyncClient:
            return self

        def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc: BaseException | None,
            traceback: Any,
        ) -> None:
            del exc_type, exc, traceback

        def post(self, url: str, *, content: str, headers: dict[str, str]) -> httpx.Response:
            seen_calls.append((url, content, headers))
            return httpx.Response(
                status_code=200,
                request=httpx.Request("POST", url, content=content, headers=headers),
                json={
                    "access_token": "sync_refreshed_access",
                    "refresh_token": "sync_refreshed_refresh",
                },
            )

    monkeypatch.setattr(httpx, "Client", FakeSyncClient)
    manager = CodexTokenManager(
        config=CodexAuthConfig(auth_path=auth_path),
        store=CodexAuthStore(auth_path),
        http_client=httpx.AsyncClient(),
    )

    token = manager.get_access_token_sync()

    assert token == "sync_refreshed_access"
    assert seen_calls == [
        (
            "https://auth.openai.com/oauth/token",
            "client_id=app_EMoamEEZ73f0CkXaXp7hrann&"
            "grant_type=refresh_token&refresh_token=refresh_before",
            {"Content-Type": "application/x-www-form-urlencoded"},
        ),
    ]
    persisted_state = CodexAuthStore(auth_path).read_state()
    assert persisted_state.access_token == "sync_refreshed_access"
    assert persisted_state.refresh_token == "sync_refreshed_refresh"
    assert persisted_state.account_id == "acct_before"
    asyncio.run(manager.close())


@pytest.mark.asyncio
async def test_token_manager_refresh_fallbacks_and_response_validation(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(
        auth_path,
        access_expiry=datetime.now(tz=UTC) - timedelta(minutes=5),
        account_id="acct_before",
        refresh_token="refresh-before",
    )

    calls: list[httpx.Request] = []

    def handler(request: httpx.Request) -> httpx.Response:
        calls.append(request)
        return httpx.Response(
            200,
            json={
                "access_token": "refreshed-access",
            },
        )

    client = httpx.AsyncClient(transport=httpx.MockTransport(handler))
    manager = CodexTokenManager(
        config=CodexAuthConfig(auth_path=auth_path),
        store=CodexAuthStore(auth_path),
        http_client=client,
    )

    token = await manager.get_access_token()

    assert token == "refreshed-access"
    assert len(calls) == 1
    assert manager.current_state.refresh_token == "refresh-before"
    assert manager.current_state.account_id == "acct_before"

    assert _string_value({"token": "value"}, "token") == "value"
    assert _string_value({"token": ""}, "token") is None
    with pytest.raises(ValueError, match="return an object"):
        _response_mapping(httpx.Response(200, json=[]))

    owner_client = httpx.AsyncClient(transport=httpx.MockTransport(handler))
    owner_manager = CodexTokenManager(
        config=CodexAuthConfig(auth_path=auth_path),
        store=CodexAuthStore(auth_path),
        http_client=owner_client,
        owns_http_client=True,
    )
    await owner_manager.close()
    assert owner_client.is_closed is True

    await client.aclose()


def test_auth_config_default_path_and_refresh_deadline_without_expiry(
    tmp_path: Path,
) -> None:
    auth_path = tmp_path / "auth.json"
    write_auth_file(auth_path, account_id="acct_deadline")
    config = CodexAuthConfig(auth_path=auth_path)
    assert default_auth_path().name == "auth.json"

    state = CodexAuthState(
        access_token="access",
        refresh_token="refresh",
        last_refresh=datetime.now(tz=UTC),
    )
    client = httpx.AsyncClient()
    manager = CodexTokenManager(
        config=config,
        store=CodexAuthStore(auth_path),
        http_client=client,
    )

    try:
        manager._state = state
        deadline = manager._refresh_deadline(state)
        assert state.last_refresh is not None
        assert deadline == state.last_refresh + config.default_token_ttl

        missing_deadline_state = CodexAuthState(
            access_token="access",
            refresh_token="refresh",
        )
        assert manager._refresh_deadline(missing_deadline_state) is None
        assert manager._should_refresh(missing_deadline_state) is False
    finally:
        asyncio.run(client.aclose())



from __future__ import annotations as _annotations

import os

from codex_auth_helper import create_codex_responses_model
from pydantic_acp import run_acp
from pydantic_ai import Agent

MODEL_NAME = os.getenv("CODEX_MODEL", "gpt-5.4")

agent = Agent(
    create_codex_responses_model(
        MODEL_NAME,
        instructions=(
            "You are a concise coding assistant. "
            "Ask for clarification when the task is underspecified."
        ),
    ),
    name="codex-responses-agent",
)


def main() -> None:
    run_acp(agent=agent)


if __name__ == "__main__":
    main()



from __future__ import annotations as _annotations

from dataclasses import dataclass, field
from datetime import timedelta
from pathlib import Path

__all__ = ("CodexAuthConfig",)


def default_auth_path() -> Path:
    return Path.home() / ".codex" / "auth.json"


@dataclass(frozen=True, slots=True)
class CodexAuthConfig:
    auth_path: Path = field(default_factory=default_auth_path)
    api_base_url: str = "https://chatgpt.com/backend-api/codex"
    client_id: str = "app_EMoamEEZ73f0CkXaXp7hrann"
    default_token_ttl: timedelta = timedelta(hours=1)
    issuer: str = "https://auth.openai.com"
    refresh_margin: timedelta = timedelta(seconds=30)
    timeout_seconds: float = 30.0



from __future__ import annotations as _annotations

import argparse
import ast
import json
import urllib.error
import urllib.request
from dataclasses import dataclass
from pathlib import Path

from packaging.version import InvalidVersion, Version

PYPI_URL_TEMPLATE = "https://pypi.org/pypi/{package}/json"


@dataclass(frozen=True, slots=True)
class PackageVersionTarget:
    package_name: str
    version_file: Path


@dataclass(frozen=True, slots=True)
class PublishedVersionOverride:
    package_name: str
    version: Version


PACKAGE_TARGETS = (
    PackageVersionTarget("acpkit", Path("src/acpkit/_version.py")),
    PackageVersionTarget(
        "pydantic-acp",
        Path("packages/adapters/pydantic-acp/src/pydantic_acp/_version.py"),
    ),
    PackageVersionTarget(
        "codex-auth-helper",
        Path("packages/helpers/codex-auth-helper/src/codex_auth_helper/_version.py"),
    ),
)


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Ensure local package versions are strictly greater than the latest PyPI releases.",
    )
    parser.add_argument(
        "--published-version",
        action="append",
        default=[],
        metavar="PACKAGE=VERSION",
        help="Override the published version for a package. Useful for local validation without network access.",
    )
    return parser.parse_args()


def _read_version(version_file: Path) -> Version:
    module = ast.parse(version_file.read_text(encoding="utf-8"), filename=str(version_file))
    for statement in module.body:
        if not isinstance(statement, ast.Assign):
            continue
        if len(statement.targets) != 1:
            continue
        target = statement.targets[0]
        if not isinstance(target, ast.Name) or target.id != "__version__":
            continue
        value = statement.value
        if not isinstance(value, ast.Constant) or not isinstance(value.value, str):
            msg = f"{version_file} must assign a string literal to __version__"
            raise SystemExit(msg)
        try:
            return Version(value.value)
        except InvalidVersion as exc:
            msg = f"{version_file} contains an invalid version: {value.value!r}"
            raise SystemExit(msg) from exc
    msg = f"{version_file} does not define __version__"
    raise SystemExit(msg)


def _parse_overrides(raw_overrides: list[str]) -> dict[str, PublishedVersionOverride]:
    overrides: dict[str, PublishedVersionOverride] = {}
    for raw_override in raw_overrides:
        package_name, separator, version_text = raw_override.partition("=")
        if not separator:
            msg = f"Invalid --published-version value: {raw_override!r}"
            raise SystemExit(msg)
        normalized_package_name = package_name.strip()
        normalized_version_text = version_text.strip()
        if not normalized_package_name or not normalized_version_text:
            msg = f"Invalid --published-version value: {raw_override!r}"
            raise SystemExit(msg)
        try:
            overrides[normalized_package_name] = PublishedVersionOverride(
                package_name=normalized_package_name,
                version=Version(normalized_version_text),
            )
        except InvalidVersion as exc:
            msg = (
                f"Invalid published override for {normalized_package_name!r}: "
                f"{normalized_version_text!r}"
            )
            raise SystemExit(msg) from exc
    return overrides


def _fetch_pypi_version(package_name: str) -> Version:
    url = PYPI_URL_TEMPLATE.format(package=package_name)
    try:
        with urllib.request.urlopen(url, timeout=10) as response:
            payload = json.loads(response.read().decode("utf-8"))
    except urllib.error.HTTPError as exc:
        msg = f"Failed to fetch PyPI metadata for {package_name!r}: HTTP {exc.code}"
        raise SystemExit(msg) from exc
    except urllib.error.URLError as exc:
        msg = f"Failed to reach PyPI for {package_name!r}: {exc.reason}"
        raise SystemExit(msg) from exc

    info = payload.get("info")
    if not isinstance(info, dict):
        msg = f"PyPI metadata for {package_name!r} is missing the info object"
        raise SystemExit(msg)
    version_text = info.get("version")
    if not isinstance(version_text, str):
        msg = f"PyPI metadata for {package_name!r} is missing the published version"
        raise SystemExit(msg)
    try:
        return Version(version_text)
    except InvalidVersion as exc:
        msg = f"PyPI returned an invalid version for {package_name!r}: {version_text!r}"
        raise SystemExit(msg) from exc


def main() -> int:
    args = _parse_args()
    repo_root = Path(__file__).resolve().parents[1]
    overrides = _parse_overrides(args.published_version)
    failed = False

    for target in PACKAGE_TARGETS:
        local_version = _read_version(repo_root / target.version_file)
        published_version = (
            overrides[target.package_name].version
            if target.package_name in overrides
            else _fetch_pypi_version(target.package_name)
        )
        if local_version <= published_version:
            failed = True
            print(
                f"{target.package_name}: local version {local_version} must be greater than "
                f"published version {published_version}",
            )
            continue
        print(
            f"{target.package_name}: ok (local {local_version} > published {published_version})",
        )

    return 1 if failed else 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations as _annotations

from pydantic_acp import (
    HostAccessPolicy,
    caution_for_command,
    caution_for_path,
    format_code_block,
    format_diff_preview,
    format_terminal_status,
    single_line_summary,
    truncate_lines,
    truncate_text,
)

from .support import Path


def test_truncate_text_appends_default_marker() -> None:
    assert truncate_text("x" * 5000, limit=20).endswith("...[truncated]")


def test_truncate_lines_keeps_last_line_for_marker() -> None:
    lines = truncate_lines(["a", "b", "c", "d"], max_lines=3)

    assert lines == ["a", "b", "... [truncated]"]


def test_single_line_summary_normalizes_newlines() -> None:
    assert single_line_summary("hello\nworld", limit=20) == "hello world"


def test_format_code_block_supports_language() -> None:
    rendered = format_code_block("echo hi", language="bash")

    assert rendered == "```bash\necho hi\n```"


def test_format_diff_preview_hides_unified_headers() -> None:
    rendered = format_diff_preview("README.md", "# old\n", "# new\n")

    assert rendered.startswith("# README.md")
    assert "--- " not in rendered
    assert "+++ " not in rendered


def test_format_terminal_status_returns_expected_labels() -> None:
    assert format_terminal_status(exit_code=0, signal=None) == "ok (0)"
    assert format_terminal_status(exit_code=2, signal=None) == "fail (2)"
    assert format_terminal_status(exit_code=None, signal="SIGTERM") == "cancelled (SIGTERM)"
    assert format_terminal_status(exit_code=None, signal=None) == "running"


def test_caution_for_path_uses_host_access_policy() -> None:
    session_cwd = Path("/tmp/acpkit-projection-helpers").resolve()

    caution = caution_for_path(
        "../outside.txt",
        session_cwd=session_cwd,
        workspace_root=session_cwd,
        access_policy=HostAccessPolicy(),
    )

    assert caution is not None
    assert "workspace root" in caution


def test_caution_for_command_uses_host_access_policy() -> None:
    session_cwd = Path("/tmp/acpkit-projection-helpers").resolve()

    caution = caution_for_command(
        "python",
        args=["../scripts/build.py"],
        session_cwd=session_cwd,
        access_policy=HostAccessPolicy.strict(),
    )

    assert caution is not None
    assert "outside the active session cwd" in caution


def test_projection_helpers_cover_edge_cases_and_no_risk_paths() -> None:
    session_cwd = Path("/tmp/acpkit-projection-helpers").resolve()

    assert truncate_text("abcdef", limit=0) == ""
    assert truncate_text("abcdef", limit=3, marker="[...]") == "abc[...]"
    assert truncate_lines(["a", "b"], max_lines=0) == []
    assert truncate_lines(["a", "b"], max_lines=1) == ["... [truncated]"]
    assert single_line_summary("hello   world  again", limit=8) == "hello wo..."
    assert format_code_block("abcdef", limit=4) == "```\nabcd\n\n...[truncated]\n```"
    assert (
        format_diff_preview(
            "README.md",
            "# same\n",
            "# same\n",
            include_path_header=False,
        )
        == "(no visible changes)"
    )

    rendered = format_diff_preview(
        "README.md",
        "# old\n",
        "# new\n",
        include_path_header=False,
        include_diff_headers=True,
    )

    assert rendered.startswith("--- ")
    assert "+++ " in rendered
    assert (
        caution_for_path(
            "notes.txt",
            session_cwd=session_cwd,
            workspace_root=session_cwd,
        )
        is None
    )
    assert (
        caution_for_command(
            "python",
            args=["./script.py"],
            session_cwd=session_cwd,
            workspace_root=session_cwd,
        )
        is None
    )



from __future__ import annotations as _annotations

import asyncio
from collections.abc import Sequence
from dataclasses import replace
from typing import Any, ClassVar, Generic, TypeVar
from uuid import uuid4

from acp import PROTOCOL_VERSION
from acp.exceptions import RequestError
from acp.interfaces import Client as AcpClient
from acp.schema import (
    AgentCapabilities,
    ClientCapabilities,
    CloseSessionResponse,
    ForkSessionResponse,
    HttpMcpServer,
    Implementation,
    InitializeResponse,
    McpServerStdio,
    PlanEntry,
    PromptCapabilities,
    SessionCapabilities,
    SessionCloseCapabilities,
    SessionForkCapabilities,
    SessionListCapabilities,
    SessionResumeCapabilities,
    SseMcpServer,
)
from pydantic_ai import Agent as PydanticAgent

from ..agent_source import AgentSource
from ..bridges import PrepareToolsBridge
from ..config import AdapterConfig
from ..models import ModelOverride
from ..session.state import AcpSessionContext, JsonValue
from ._adapter_mixins import (
    _PromptRuntimeDelegationMixin,
    _SessionRuntimeDelegationMixin,
)
from ._adapter_prompt import _AdapterPromptHandler
from ._prompt_runtime import NativePlanGeneration, TaskPlan, _PromptRuntime
from ._session_runtime import _SessionRuntime
from .bridge_manager import BridgeManager
from .hook_introspection import list_agent_hooks

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
OutputDataT = TypeVar("OutputDataT", covariant=True)

__all__ = ("NativePlanGeneration", "PydanticAcpAgent", "TaskPlan")


class PydanticAcpAgent(
    _PromptRuntimeDelegationMixin[AgentDepsT, OutputDataT],
    _SessionRuntimeDelegationMixin[AgentDepsT, OutputDataT],
    Generic[AgentDepsT, OutputDataT],
):
    """Expose a `pydantic_ai.Agent` as an ACP-compatible session runtime."""

    _pydantic_acp_meta_supported: ClassVar[bool] = True

    def __init__(
        self,
        agent_source: AgentSource[AgentDepsT, OutputDataT],
        *,
        config: AdapterConfig,
    ) -> None:
        self._agent_source = agent_source
        bridge_projection_maps = tuple(
            projection_map
            for bridge in config.capability_bridges
            for projection_map in bridge.get_projection_maps()
        )
        self._config = (
            replace(
                config,
                projection_maps=(*config.projection_maps, *bridge_projection_maps),
            )
            if bridge_projection_maps
            else config
        )
        self._client: AcpClient | None = None
        self._bridge_manager = BridgeManager(
            base_classifier=self._config.tool_classifier,
            bridges=tuple(self._config.capability_bridges),
        )
        self._tool_classifier = self._bridge_manager.tool_classifier
        self._prompt_runtime = _PromptRuntime(self)
        self._adapter_prompt = _AdapterPromptHandler(self)
        self._session_runtime = _SessionRuntime(self)
        self._active_prompt_tasks: dict[str, asyncio.Task[Any]] = {}

    def on_connect(self, conn: AcpClient) -> None:
        self._client = conn

    def _new_session_id(self) -> str:
        return uuid4().hex

    def _list_agent_hooks(self, agent: PydanticAgent[AgentDepsT, OutputDataT]) -> list[Any]:
        return list_agent_hooks(agent)

    async def initialize(
        self,
        protocol_version: int,
        client_capabilities: ClientCapabilities | None = None,
        client_info: Implementation | None = None,
        **kwargs: Any,
    ) -> InitializeResponse:
        """Negotiate ACP protocol version and advertise adapter capabilities."""
        del client_capabilities, client_info, kwargs
        negotiated_version = min(protocol_version, PROTOCOL_VERSION)
        return InitializeResponse(
            protocol_version=negotiated_version,
            agent_capabilities=AgentCapabilities(
                load_session=True,
                mcp_capabilities=self._bridge_manager.get_mcp_capabilities(),
                prompt_capabilities=PromptCapabilities(
                    audio=self._config.prompt_capabilities.audio,
                    embedded_context=self._config.prompt_capabilities.embedded_context,
                    image=self._config.prompt_capabilities.image,
                ),
                session_capabilities=SessionCapabilities(
                    close=SessionCloseCapabilities(),
                    fork=SessionForkCapabilities(),
                    list=SessionListCapabilities(),
                    resume=SessionResumeCapabilities(),
                ),
            ),
            agent_info=Implementation(
                name=self._config.agent_name,
                title=self._config.agent_title,
                version=self._config.agent_version,
            ),
        )

    async def authenticate(self, method_id: str, **kwargs: Any) -> None:
        """Accept ACP auth handshakes when the host does not require extra auth."""
        del method_id, kwargs

    async def fork_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None = None,
        **kwargs: Any,
    ) -> ForkSessionResponse:
        del kwargs
        response = await self._session_runtime.fork_session(cwd, session_id, mcp_servers)
        return ForkSessionResponse(
            session_id=response.session_id,
            config_options=response.config_options,
            models=response.models,
            modes=response.modes,
        )

    async def close_session(self, session_id: str, **kwargs: Any) -> CloseSessionResponse | None:
        """Close a persisted ACP session if it exists."""
        del kwargs
        if not await self._session_runtime.close_session(session_id):
            return None
        return CloseSessionResponse()

    async def cancel(self, session_id: str, **kwargs: Any) -> None:
        """Cancel the active prompt task for a session, if one is running."""
        del kwargs
        active_task = self._active_prompt_tasks.get(session_id)
        if active_task is not None and not active_task.done():
            active_task.cancel()

    async def ext_method(self, method: str, params: dict[str, Any]) -> dict[str, Any]:
        """Reject unsupported ACP extension methods."""
        del params
        raise RequestError.method_not_found(method)

    async def ext_notification(self, method: str, params: dict[str, Any]) -> None:
        """Ignore unsupported ACP extension notifications."""
        del method, params

    def _native_plan_bridge(
        self,
        session: AcpSessionContext,
    ) -> PrepareToolsBridge[Any] | None:
        return self._prompt_runtime._native_plan_bridge(session)

    def _supports_native_plan_state(self, session: AcpSessionContext) -> bool:
        return self._prompt_runtime._supports_native_plan_state(session)

    def _get_native_plan_entries(self, session: AcpSessionContext) -> list[PlanEntry] | None:
        return self._prompt_runtime._get_native_plan_entries(session)

    def _set_native_plan_state(
        self,
        session: AcpSessionContext,
        *,
        entries: Sequence[PlanEntry],
        plan_markdown: str | None,
    ) -> None:
        self._prompt_runtime._set_native_plan_state(
            session,
            entries=entries,
            plan_markdown=plan_markdown,
        )

    def _format_native_plan(self, session: AcpSessionContext) -> str:
        return self._prompt_runtime._format_native_plan(session)

    async def _persist_current_native_plan_state(
        self,
        session: AcpSessionContext,
        *,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        await self._prompt_runtime._persist_current_native_plan_state(
            session,
            agent=agent,
        )

    def _install_native_plan_tools(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        self._prompt_runtime._install_native_plan_tools(agent)

    def _consume_native_plan_update(
        self,
        session: AcpSessionContext,
    ) -> bool:
        return self._prompt_runtime._consume_native_plan_update(session)

    def _model_identity(self, model_value: ModelOverride | None) -> str | None:
        return self._session_runtime._model_identity(model_value)

    async def _handle_slash_command(
        self,
        command_name: str,
        *,
        argument: str | None,
        session: AcpSessionContext,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> str | None:
        return await self._session_runtime._handle_slash_command(
            command_name,
            argument=argument,
            session=session,
            agent=agent,
        )

    def _apply_session_model_to_agent(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        session: AcpSessionContext,
    ) -> None:
        self._session_runtime._apply_session_model_to_agent(agent, session)

    def _set_agent_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        session: AcpSessionContext,
        model_id: str,
    ) -> None:
        self._session_runtime._set_agent_model(agent, session, model_id)

    def _restore_default_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
        session: AcpSessionContext,
    ) -> bool:
        return self._session_runtime._restore_default_model(agent, session)

    def _restore_agent_default_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> bool:
        return self._session_runtime._restore_agent_default_model(agent)

    def _remember_default_model(
        self,
        agent: PydanticAgent[AgentDepsT, OutputDataT],
    ) -> None:
        self._session_runtime._remember_default_model(agent)

    def _resolve_selected_model(self, model_id: str) -> ModelOverride:
        return self._session_runtime._resolve_selected_model(model_id)

    def _update_session_mcp_servers(
        self,
        session: AcpSessionContext,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None,
    ) -> None:
        self._session_runtime._update_session_mcp_servers(session, mcp_servers)

    def _serialize_mcp_server(
        self,
        server: HttpMcpServer | McpServerStdio | SseMcpServer,
    ) -> dict[str, JsonValue]:
        return self._session_runtime._serialize_mcp_server(server)

    def _require_session(self, session_id: str) -> AcpSessionContext:
        return self._session_runtime._require_session(session_id)



from __future__ import annotations as _annotations

import contextlib
import json
import os
from dataclasses import dataclass
from json import JSONDecodeError
from pathlib import Path
from uuid import uuid4

from .state import CodexAuthState

__all__ = ("CodexAuthStore",)


@dataclass(slots=True)
class CodexAuthStore:
    path: Path

    def read_state(self) -> CodexAuthState:
        try:
            text = self.path.read_text(encoding="utf-8")
        except FileNotFoundError as exc:
            raise FileNotFoundError(f"Codex auth file was not found at `{self.path}`.") from exc

        try:
            raw = json.loads(text)
        except JSONDecodeError as exc:
            raise ValueError(
                f"Codex auth file at `{self.path}` does not contain valid JSON.",
            ) from exc

        if not isinstance(raw, dict):
            raise ValueError(f"Codex auth file at `{self.path}` must contain a JSON object.")
        return CodexAuthState.from_json_dict(raw)

    def write_state(self, state: CodexAuthState) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        encoded = json.dumps(state.to_json_dict(), indent=2) + "\n"
        temp_path = self.path.with_name(f".{self.path.name}.{uuid4().hex}.tmp")
        try:
            file_descriptor = os.open(temp_path, os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0o600)
            with os.fdopen(file_descriptor, "w", encoding="utf-8") as temp_file:
                temp_file.write(encoded)
                temp_file.flush()
                os.fsync(temp_file.fileno())
            os.replace(temp_path, self.path)
            with contextlib.suppress(OSError):
                os.chmod(self.path, 0o600)
            _fsync_directory(self.path.parent)
        finally:
            temp_path.unlink(missing_ok=True)


def _fsync_directory(path: Path) -> None:
    try:
        directory_fd = os.open(path, os.O_RDONLY)
    except OSError:
        return
    try:
        with contextlib.suppress(OSError):
            os.fsync(directory_fd)
    finally:
        os.close(directory_fd)



from __future__ import annotations as _annotations

from types import SimpleNamespace
from typing import Any, cast

from .support import (
    UTC,
    AcpSessionContext,
    AdapterConfig,
    Agent,
    AgentBridgeBuilder,
    ApprovalRequired,
    McpBridge,
    McpServerDefinition,
    McpToolDefinition,
    MemorySessionStore,
    NativeApprovalBridge,
    Path,
    RecordingClient,
    RunContext,
    SessionConfigOptionBoolean,
    SessionConfigOptionSelect,
    SessionConfigSelectGroup,
    SessionConfigSelectOption,
    SessionInfoUpdate,
    SessionMcpBridge,
    TestModel,
    create_acp_agent,
    datetime,
    text_block,
)


def test_mcp_bridge_handles_empty_and_prefix_scoped_behaviour() -> None:
    session = AcpSessionContext(
        session_id="session-mcp",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    agent = Agent(TestModel(custom_output_text="unused"))

    empty_bridge = McpBridge()
    assert empty_bridge.get_mcp_capabilities() is None
    assert empty_bridge.get_session_metadata(session, agent) is None
    assert empty_bridge.get_config_options(session, agent) is None
    assert empty_bridge.get_tool_kind("missing") is None
    assert empty_bridge.get_approval_policy_key("missing") is None

    bridge = McpBridge(
        approval_policy_scope="prefix",
        servers=[
            McpServerDefinition(
                server_id="repo",
                name="Repo",
                transport="http",
                url="https://repo.example/mcp",
                tool_prefix="mcp.repo.",
            ),
        ],
        tools=[
            McpToolDefinition(
                tool_name="mcp.explicit.search",
                server_id="repo",
                kind="search",
            ),
        ],
        config_options=[
            SessionConfigOptionBoolean(
                id="mcp_enabled",
                name="MCP Enabled",
                current_value=False,
                type="boolean",
            ),
            SessionConfigOptionSelect(
                id="mcp_scope",
                name="MCP Scope",
                current_value="repo",
                options=[
                    SessionConfigSelectOption(name="Repo", value="repo"),
                    SessionConfigSelectOption(name="Docs", value="docs"),
                ],
                type="select",
            ),
        ],
    )

    capabilities = bridge.get_mcp_capabilities()
    assert capabilities is not None
    assert capabilities.http is True
    assert bridge.get_tool_kind("mcp.explicit.search") == "search"
    assert bridge.get_tool_kind("mcp.repo.read_file") == "execute"
    assert bridge.get_approval_policy_key("mcp.explicit.search") == "mcp:prefix:mcp.repo."
    assert bridge.get_approval_policy_key("mcp.repo.read_file") == "mcp:prefix:mcp.repo."
    assert bridge.get_approval_policy_key("missing.tool") is None

    assert bridge.set_config_option(session, agent, "mcp_enabled", "yes") is None
    assert bridge.set_config_option(session, agent, "mcp_scope", True) is None
    assert bridge.set_config_option(session, agent, "mcp_scope", "invalid") is None

    updated_options = bridge.set_config_option(session, agent, "mcp_enabled", True)
    assert updated_options is not None
    updated_options = bridge.set_config_option(session, agent, "mcp_scope", "docs")
    assert updated_options is not None

    metadata = bridge.get_session_metadata(session, agent)
    assert metadata is not None
    assert metadata["approval_policy_scope"] == "prefix"
    assert metadata["config_option_ids"] == ["mcp_enabled", "mcp_scope"]
    assert metadata["config"] == {"mcp_enabled": True, "mcp_scope": "docs"}


def test_session_mcp_bridge_builds_toolset_from_session_servers() -> None:
    session = AcpSessionContext(
        session_id="session-mcp-toolset",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
        mcp_servers=[
            {
                "headers": {"Authorization": "Bearer secret"},
                "name": "repo",
                "transport": "http",
                "url": "https://repo.example/mcp",
            },
            {
                "args": ["server.py"],
                "command": "python",
                "env": {"TOKEN": "secret"},
                "name": "repo",
                "transport": "stdio",
            },
        ],
    )
    bridge = SessionMcpBridge(
        include_return_schema=True,
        tool_name_prefixes=frozenset({"repo_"}),
    )

    capabilities = bridge.build_agent_capabilities(session)

    assert len(capabilities) == 1
    assert (
        bridge.build_agent_capabilities(
            AcpSessionContext(
                session_id="empty",
                cwd=Path("/tmp"),
                created_at=datetime.now(UTC),
                updated_at=datetime.now(UTC),
            ),
        )
        == ()
    )
    assert bridge.get_tool_kind("repo_search") == "execute"
    assert bridge.get_tool_kind("search") is None
    assert bridge.get_mcp_capabilities() is not None
    assert (
        SessionMcpBridge(advertise_http=False, advertise_sse=False).get_mcp_capabilities() is None
    )

    capability = cast("Any", capabilities[0])
    toolset = capability.local
    client = toolset.client
    config = client.transport.config
    servers = config.mcpServers

    assert set(servers) == {"repo", "repo-2"}
    assert servers["repo"].url == "https://repo.example/mcp"
    assert servers["repo"].headers == {"Authorization": "Bearer secret"}
    assert servers["repo-2"].command == "python"
    assert servers["repo-2"].args == ["server.py"]
    assert servers["repo-2"].env == {"TOKEN": "secret"}

    metadata = bridge.get_session_metadata(session, Agent(TestModel()))
    assert metadata == {
        "allowed_tools": None,
        "cache_prompts": True,
        "cache_resources": True,
        "cache_tools": True,
        "include_instructions": True,
        "include_return_schema": True,
        "server_count": 2,
        "servers": [
            {
                "header_names": ["Authorization"],
                "name": "repo",
                "transport": "http",
                "url": "https://repo.example/mcp",
            },
            {
                "args": ["server.py"],
                "command": "python",
                "env_names": ["TOKEN"],
                "name": "repo",
                "transport": "stdio",
            },
        ],
        "tool_error_behavior": "retry",
        "tool_name_prefixes": ["repo_"],
    }


def test_mcp_bridge_tool_scope_and_config_only_metadata() -> None:
    session = AcpSessionContext(
        session_id="session-mcp-tool",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    agent = Agent(TestModel(custom_output_text="unused"))
    bridge = McpBridge(
        approval_policy_scope="tool",
        servers=[
            McpServerDefinition(
                server_id="docs",
                name="Docs",
                transport="sse",
            ),
        ],
        config_options=[
            SessionConfigOptionBoolean(
                id="enabled",
                name="Enabled",
                current_value=True,
                type="boolean",
            ),
        ],
    )

    capabilities = bridge.get_mcp_capabilities()
    assert capabilities is not None
    assert capabilities.http is False
    assert capabilities.sse is True
    assert bridge.get_approval_policy_key("missing.tool") is None
    metadata = bridge.get_session_metadata(session, agent)
    assert metadata is not None
    assert "servers" in metadata
    assert metadata["config"] == {"enabled": True}

    config_only_bridge = McpBridge(
        config_options=[
            SessionConfigOptionBoolean(
                id="config_only",
                name="Config Only",
                current_value=False,
                type="boolean",
            ),
        ],
    )
    config_only_metadata = config_only_bridge.get_session_metadata(session, agent)
    assert config_only_metadata == {
        "approval_policy_scope": "tool",
        "config_option_ids": ["config_only"],
        "config": {"config_only": False},
    }
    assert config_only_bridge.get_tool_kind("unknown.tool") is None

    server_scope_bridge = McpBridge(
        approval_policy_scope="server",
        servers=[
            McpServerDefinition(
                server_id="docs",
                name="Docs",
                transport="http",
                tool_prefix="docs.",
            ),
        ],
    )
    assert server_scope_bridge.get_approval_policy_key("docs.search") == "mcp:server:docs"

    prefix_fallback_bridge = McpBridge(
        approval_policy_scope="prefix",
        tools=[McpToolDefinition(tool_name="plain.tool", server_id="missing", kind="execute")],
    )
    assert prefix_fallback_bridge.get_approval_policy_key("plain.tool") == "mcp:tool:plain.tool"
    assert prefix_fallback_bridge._find_config_option("missing") is None
    assert prefix_fallback_bridge._find_server("missing") is None
    assert prefix_fallback_bridge._sync_config_option(
        cast("Any", SimpleNamespace(id="other")),
        session,
    ) == cast("Any", SimpleNamespace(id="other"))

    select_bridge = McpBridge(
        config_options=[
            SessionConfigOptionSelect(
                id="scope",
                name="Scope",
                current_value="repo",
                options=[
                    SessionConfigSelectOption(name="Repo", value="repo"),
                    SessionConfigSelectOption(name="Docs", value="docs"),
                ],
                type="select",
            ),
        ],
        servers=[
            McpServerDefinition(
                server_id="alpha",
                name="Alpha",
                transport="http",
                tool_prefix=None,
            ),
            McpServerDefinition(
                server_id="docs",
                name="Docs",
                transport="http",
                tool_prefix="docs.",
            ),
        ],
    )
    assert select_bridge.get_tool_kind("plain") is None
    assert select_bridge.set_config_option(session, agent, "missing", True) is None
    assert select_bridge.set_config_option(session, agent, "scope", "docs") is not None
    assert select_bridge._find_server("docs") is not None
    assert select_bridge._find_server_for_tool("docs.search") is not None
    synced = select_bridge._sync_config_option(
        select_bridge.config_options[0],
        session,
    )
    assert synced.current_value == "docs"
    custom_bridge = McpBridge(config_options=[cast("Any", SimpleNamespace(id="custom"))])
    custom_options = custom_bridge.set_config_option(session, agent, "custom", "value")
    assert custom_options is not None
    assert session.config_values["custom"] == "value"


def test_mcp_bridge_accepts_grouped_select_options() -> None:
    session = AcpSessionContext(
        session_id="session-mcp-groups",
        cwd=Path("/tmp"),
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    agent = Agent(TestModel(custom_output_text="unused"))
    bridge = McpBridge(
        config_options=[
            SessionConfigOptionSelect(
                id="scope",
                name="Scope",
                current_value="repo",
                options=[
                    SessionConfigSelectGroup(
                        group="workspace",
                        name="Workspace",
                        options=[
                            SessionConfigSelectOption(name="Repo", value="repo"),
                            SessionConfigSelectOption(name="Docs", value="docs"),
                        ],
                    ),
                    SessionConfigSelectGroup(
                        group="external",
                        name="External",
                        options=[SessionConfigSelectOption(name="Web", value="web")],
                    ),
                ],
                type="select",
            ),
        ],
    )

    assert bridge.set_config_option(session, agent, "scope", "docs") is not None
    assert bridge.set_config_option(session, agent, "scope", "web") is not None
    assert bridge.set_config_option(session, agent, "scope", "missing") is None


async def test_mcp_bridge_exposes_config_and_routes_server_scoped_approval(
    tmp_path: Path,
) -> None:
    mcp_bridge = McpBridge(
        approval_policy_scope="server",
        config_options=[
            SessionConfigOptionBoolean(
                id="mcp_auto_connect",
                name="Auto Connect",
                category="mcp",
                description="Connect MCP tools automatically.",
                type="boolean",
                current_value=False,
            ),
        ],
        servers=[
            McpServerDefinition(
                server_id="repo",
                name="Repo MCP",
                transport="http",
                tool_prefix="mcp.repo.",
            ),
        ],
        tools=[
            McpToolDefinition(tool_name="mcp.repo.alpha", server_id="repo", kind="read"),
            McpToolDefinition(tool_name="mcp.repo.beta", server_id="repo", kind="read"),
        ],
    )

    def factory(session: AcpSessionContext) -> Agent[None, str]:
        builder = AgentBridgeBuilder(session=session, capability_bridges=[mcp_bridge])
        contributions = builder.build()
        tool_name = "mcp.repo.alpha" if len(session.transcript) <= 1 else "mcp.repo.beta"
        agent = Agent(
            TestModel(call_tools=[tool_name]),
            capabilities=contributions.capabilities,
        )

        @agent.tool(name="mcp.repo.alpha")
        def repo_alpha(ctx: RunContext[None], path: str) -> str:
            if not ctx.tool_call_approved:
                raise ApprovalRequired()
            return f"alpha:{path}"

        @agent.tool(name="mcp.repo.beta")
        def repo_beta(ctx: RunContext[None], path: str) -> str:
            if not ctx.tool_call_approved:  # pragma: no cover
                raise ApprovalRequired()  # pragma: no cover
            return f"beta:{path}"  # pragma: no cover

        return agent

    adapter = create_acp_agent(
        agent_factory=factory,
        config=AdapterConfig(
            approval_bridge=NativeApprovalBridge(enable_persistent_choices=True),
            capability_bridges=[mcp_bridge],
            session_store=MemorySessionStore(),
        ),
    )
    client = RecordingClient()
    client.queue_permission_selected("allow_always")
    adapter.on_connect(client)

    session = await adapter.new_session(cwd=str(tmp_path), mcp_servers=[])
    assert session.config_options is not None
    auto_connect_option = next(
        option for option in session.config_options if option.id == "mcp_auto_connect"
    )
    assert auto_connect_option.current_value is False

    config_response = await adapter.set_config_option(
        config_id="mcp_auto_connect",
        session_id=session.session_id,
        value=True,
    )

    assert config_response is not None
    updated_auto_connect_option = next(
        option for option in config_response.config_options if option.id == "mcp_auto_connect"
    )
    assert updated_auto_connect_option.current_value is True

    first_prompt = await adapter.prompt(
        prompt=[text_block("Use the first MCP tool.")],
        session_id=session.session_id,
    )
    second_prompt = await adapter.prompt(
        prompt=[text_block("Use the second MCP tool.")],
        session_id=session.session_id,
    )

    assert first_prompt.stop_reason == "end_turn"
    assert second_prompt.stop_reason == "end_turn"
    assert len(client.permission_option_ids) == 1
    assert client.permission_option_ids[0][1] == [
        "allow_once",
        "allow_always",
        "reject_once",
        "reject_always",
    ]
    assert client.permission_option_ids[0][2].kind == "read"

    session_info_updates = [
        update for _, update in client.updates if isinstance(update, SessionInfoUpdate)
    ]
    assert session_info_updates
    session_info = session_info_updates[-1]
    assert session_info.field_meta is not None
    assert session_info.field_meta["pydantic_acp"]["mcp"] == {
        "approval_policy_scope": "server",
        "config": {"mcp_auto_connect": True},
        "config_option_ids": ["mcp_auto_connect"],
        "servers": [
            {
                "description": None,
                "name": "Repo MCP",
                "server_id": "repo",
                "tool_prefix": "mcp.repo.",
                "transport": "http",
                "url": None,
            },
        ],
    }



from __future__ import annotations as _annotations

import asyncio

import pytest
from pydantic_acp import HostAccessPolicy
from pydantic_acp.host._policy_commands import (
    extract_command_path_candidates,
    resolve_command_cwd,
)

from .support import (
    UTC,
    AcpSessionContext,
    ClientHostContext,
    FilesystemRecordingClient,
    Path,
    TerminalRecordingClient,
    datetime,
)


def test_host_access_policy_warns_for_absolute_path_inside_workspace() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    policy = HostAccessPolicy()

    evaluation = policy.evaluate_path(
        session_cwd / "notes.txt",
        session_cwd=session_cwd,
        workspace_root=session_cwd,
    )

    assert evaluation.disposition == "warn"
    assert evaluation.is_absolute_input is True
    assert evaluation.outside_cwd is False
    assert evaluation.outside_workspace is False
    assert evaluation.should_warn is True
    assert evaluation.should_deny is False
    assert evaluation.headline == "Host access requires review"
    assert evaluation.recommendation == "Review the target carefully before continuing."
    assert evaluation.risk_codes == ("absolute_path",)
    assert "Path is absolute" in evaluation.message


def test_host_access_policy_denies_path_outside_workspace() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    policy = HostAccessPolicy()

    evaluation = policy.evaluate_path(
        "../outside.txt",
        session_cwd=session_cwd,
        workspace_root=session_cwd,
    )

    assert evaluation.disposition == "deny"
    assert evaluation.outside_cwd is True
    assert evaluation.outside_workspace is True
    assert "Path escapes the workspace root" in evaluation.message


def test_host_access_policy_warns_for_command_target_outside_cwd_without_workspace_root() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    policy = HostAccessPolicy()

    evaluation = policy.evaluate_command(
        "python",
        args=["../tools/run.py"],
        session_cwd=session_cwd,
    )

    assert evaluation.disposition == "warn"
    assert evaluation.references_external_paths is True
    assert evaluation.references_outside_workspace is False
    assert evaluation.primary_risk is not None
    assert evaluation.summary_lines()[0] == "Host access requires review"
    assert "outside the active session cwd" in evaluation.message


def test_host_access_policy_strict_preset_denies_external_command_paths() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    policy = HostAccessPolicy.strict()

    evaluation = policy.evaluate_command(
        "python",
        args=["../tools/run.py"],
        session_cwd=session_cwd,
    )

    assert evaluation.disposition == "deny"
    assert evaluation.should_deny is True
    assert evaluation.headline == "Host access denied"
    assert evaluation.recommendation == "Do not continue unless you intentionally relax the policy."


def test_host_access_policy_permissive_preset_warns_for_outside_workspace() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    policy = HostAccessPolicy.permissive()

    evaluation = policy.evaluate_path(
        "../outside.txt",
        session_cwd=session_cwd,
        workspace_root=session_cwd,
    )

    assert evaluation.disposition == "warn"
    assert evaluation.should_warn is True
    assert evaluation.headline == "Host access requires review"


def test_host_access_policy_allows_safe_relative_paths_and_commands() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    policy = HostAccessPolicy()

    path_evaluation = policy.enforce_path(
        "notes.txt",
        session_cwd=session_cwd,
        workspace_root=session_cwd,
    )
    command_evaluation = policy.enforce_command(
        "python",
        args=["./script.py"],
        session_cwd=session_cwd,
        workspace_root=session_cwd,
    )

    assert path_evaluation.disposition == "allow"
    assert path_evaluation.has_risks is False
    assert path_evaluation.primary_risk is None
    assert path_evaluation.message == "Host access allowed."
    assert path_evaluation.summary_lines() == (
        "Host access allowed",
        "Host access allowed.",
        "Safe to continue.",
    )
    assert command_evaluation.disposition == "allow"
    assert command_evaluation.has_risks is False
    assert command_evaluation.primary_risk is None
    assert command_evaluation.message == "Host access allowed."
    assert command_evaluation.referenced_paths == (
        (session_cwd / "script.py").resolve(strict=False),
    )
    assert command_evaluation.summary_lines() == (
        "Host access allowed",
        "Host access allowed.",
        "Safe to continue.",
    )


def test_host_access_policy_warns_for_command_cwd_outside_workspace() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    policy = HostAccessPolicy.permissive()

    evaluation = policy.evaluate_command(
        "python",
        cwd="../outside",
        session_cwd=session_cwd,
        workspace_root=session_cwd,
    )

    assert evaluation.disposition == "warn"
    assert evaluation.should_warn is True
    assert evaluation.outside_cwd is True
    assert evaluation.outside_workspace is True
    assert evaluation.risk_codes == (
        "command_cwd_outside_cwd",
        "command_cwd_outside_workspace",
    )
    assert "Command cwd escapes the workspace root" in evaluation.message


def test_command_path_helpers_extract_relative_flag_and_assignment_tokens() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()

    assert resolve_command_cwd(None, session_cwd=session_cwd) == session_cwd
    assert resolve_command_cwd("../runner", session_cwd=session_cwd) == (
        session_cwd / "../runner"
    ).resolve(strict=False)
    assert extract_command_path_candidates(
        "./bin/run.py",
        args=[
            "--output=../dist/result.txt",
            "cwd=..",
            "./bin/run.py",
            ".",
            "..",
            "--name=value",
            "name=value",
            "",
            "--flag",
            "not-a-path",
        ],
        command_cwd=session_cwd,
    ) == (
        (session_cwd / "bin/run.py").resolve(strict=False),
        (session_cwd / "../dist/result.txt").resolve(strict=False),
        session_cwd.parent.resolve(strict=False),
        session_cwd,
    )


def test_client_filesystem_backend_enforces_denying_policy() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    session = AcpSessionContext(
        session_id="session-host-policy-filesystem",
        cwd=session_cwd,
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    client = FilesystemRecordingClient()
    host = ClientHostContext.from_session(
        client=client,
        session=session,
        access_policy=HostAccessPolicy(),
        workspace_root=session_cwd,
    )

    with pytest.raises(PermissionError, match="workspace root"):
        asyncio.run(host.filesystem.write_text_file("../outside.txt", "blocked"))

    assert client.write_calls == []


def test_client_terminal_backend_enforces_denying_policy() -> None:
    session_cwd = Path("/tmp/acpkit-host-policy").resolve()
    session = AcpSessionContext(
        session_id="session-host-policy-terminal",
        cwd=session_cwd,
        created_at=datetime.now(UTC),
        updated_at=datetime.now(UTC),
    )
    client = TerminalRecordingClient()
    host = ClientHostContext.from_session(
        client=client,
        session=session,
        access_policy=HostAccessPolicy(),
        workspace_root=session_cwd,
    )

    with pytest.raises(PermissionError, match="workspace root"):
        asyncio.run(
            host.terminal.create_terminal(
                "python",
                args=["/tmp/outside.py"],
            ),
        )

    assert client.create_calls == []



from __future__ import annotations as _annotations

from typing import TYPE_CHECKING, Generic, TypeVar

from acp.exceptions import RequestError
from acp.schema import (
    HttpMcpServer,
    McpServerStdio,
    NewSessionResponse,
    ResumeSessionResponse,
    SseMcpServer,
)

from ..session.state import AcpSessionContext, utc_now

if TYPE_CHECKING:
    from ._session_runtime import _SessionRuntime
    from .session_surface import SessionSurface

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
OutputDataT = TypeVar("OutputDataT", covariant=True)

__all__ = ("_SessionLifecycle",)


class _SessionLifecycle(Generic[AgentDepsT, OutputDataT]):
    def __init__(self, runtime: _SessionRuntime[AgentDepsT, OutputDataT]) -> None:
        self._runtime = runtime

    async def new_session(
        self,
        cwd: str,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None = None,
    ) -> NewSessionResponse:
        owner = self._runtime._owner
        session = self._runtime._bind_session_client(
            AcpSessionContext(
                session_id=owner._new_session_id(),
                cwd=self._runtime._normalize_cwd(cwd),
                created_at=utc_now(),
                updated_at=utc_now(),
            ),
        )
        self._runtime._update_session_mcp_servers(session, mcp_servers)
        owner._config.session_store.save(session)
        surface = await self._prepare_session_surface(session, replay_transcript=False)
        return self._new_session_response(session, surface=surface)

    async def load_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None = None,
    ) -> NewSessionResponse | None:
        session = self._runtime._owner._config.session_store.get(session_id)
        if session is None:
            return None
        session = self._runtime._bind_session_client(session)
        session.cwd = self._runtime._normalize_cwd(cwd)
        self._runtime._update_session_mcp_servers(session, mcp_servers)
        session.updated_at = utc_now()
        self._runtime._owner._config.session_store.save(session)
        surface = await self._prepare_session_surface(session, replay_transcript=True)
        return self._new_session_response(session, surface=surface)

    async def fork_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None = None,
    ) -> NewSessionResponse:
        owner = self._runtime._owner
        forked_session = owner._config.session_store.fork(
            session_id,
            new_session_id=owner._new_session_id(),
            cwd=self._runtime._normalize_cwd(cwd),
        )
        if forked_session is None:
            raise RequestError.invalid_params({"sessionId": session_id})
        forked_session = self._runtime._bind_session_client(forked_session)
        self._runtime._update_session_mcp_servers(forked_session, mcp_servers)
        surface = await self._prepare_session_surface(forked_session, replay_transcript=False)
        return self._new_session_response(forked_session, surface=surface)

    async def resume_session(
        self,
        cwd: str,
        session_id: str,
        mcp_servers: list[HttpMcpServer | McpServerStdio | SseMcpServer] | None = None,
    ) -> ResumeSessionResponse:
        session = self._runtime._require_session(session_id)
        session.cwd = self._runtime._normalize_cwd(cwd)
        self._runtime._update_session_mcp_servers(session, mcp_servers)
        session.updated_at = utc_now()
        self._runtime._owner._config.session_store.save(session)
        surface = await self._prepare_session_surface(session, replay_transcript=True)
        return ResumeSessionResponse(
            config_options=surface.config_options,
            models=surface.model_state,
            modes=surface.mode_state,
        )

    async def _prepare_session_surface(
        self,
        session: AcpSessionContext,
        *,
        replay_transcript: bool,
    ) -> SessionSurface:
        owner = self._runtime._owner
        if replay_transcript:
            await owner._replay_transcript(session)
        agent = await owner._agent_source.get_agent(session)
        self._runtime._configure_agent_runtime(session, agent)
        surface = await self._runtime._build_session_surface(session, agent)
        await self._runtime._emit_session_state_updates(
            session,
            surface,
            emit_available_commands=True,
            emit_config_options=False,
            emit_current_mode=False,
            emit_plan=True,
            emit_session_info=True,
        )
        return surface

    def _new_session_response(
        self,
        session: AcpSessionContext,
        *,
        surface: SessionSurface,
    ) -> NewSessionResponse:
        return NewSessionResponse(
            session_id=session.session_id,
            config_options=surface.config_options,
            models=surface.model_state,
            modes=surface.mode_state,
        )



from __future__ import annotations as _annotations

import argparse
import os
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Final

from pydantic_acp import (
    AcpSessionContext,
    AdapterConfig,
    AgentBridgeBuilder,
    AgentFactory,
    CapabilityBridge,
    FileSessionStore,
    HarnessCodeModeBridge,
    HarnessFileSystemBridge,
    HarnessShellBridge,
    create_acp_agent,
    run_acp,
)
from pydantic_ai import Agent

if TYPE_CHECKING:
    from codex_auth_helper import CodexResponsesModel

__all__ = ("acp_agent", "agent_factory", "config", "main")

_AGENT_NAME: Final[str] = "harness-agent"
_DEFAULT_MODEL_NAME: Final[str] = "openrouter:google/gemini-3-flash-preview"
_DEFAULT_CODEX_MODEL: Final[str] = "gpt-5.4"
_DEMO_ROOT: Final[Path] = Path("agent_demos")
_WORKSPACE_ROOT: Final[Path] = Path.cwd() / _DEMO_ROOT / "harness-agent"
_SESSION_STORE_ROOT: Final[Path] = (
    Path(os.getenv("ACP_EXAMPLE_SESSION_DIR", str(_DEMO_ROOT / "acp-sessions")))
    .expanduser()
    .resolve()
    / "pydantic-harness"
)
_INSTRUCTIONS: Final[str] = (
    "You are an ACP Kit harness agent. Use the pydantic-ai-harness filesystem and shell "
    "tools for concrete workspace work. Keep all filesystem and shell work inside the "
    "seeded harness workspace. Prefer inspecting existing files before editing or running "
    "commands, and summarize tool results plainly."
)
_CODE_MODE_INSTRUCTIONS: Final[str] = (
    " Code-mode tools are enabled for this run; use them only when Python execution or "
    "multi-step code inspection is actually useful."
)
_DEFAULT_FILES: Final[dict[str, str]] = {
    "README.md": (
        "# Harness Workspace\n\n"
        "Use the harness file and shell tools for real workspace operations here. "
        "Run this example with `--codemode` to enable code-mode tools.\n"
    ),
    "notes/todo.md": "- inspect README.md\n- run a harmless shell command\n- explain findings\n",
}


def _ensure_workspace() -> None:
    _WORKSPACE_ROOT.mkdir(parents=True, exist_ok=True)
    for relative_path, content in _DEFAULT_FILES.items():
        file_path = _WORKSPACE_ROOT / relative_path
        file_path.parent.mkdir(parents=True, exist_ok=True)
        if not file_path.exists():
            file_path.write_text(content, encoding="utf-8")


def _build_harness_bridges(
    root: Path,
    *,
    include_code_mode: bool = False,
) -> tuple[CapabilityBridge, ...]:
    bridges: list[CapabilityBridge] = [
        HarnessFileSystemBridge(
            root_dir=root,
            allowed_patterns=("**/*",),
            protected_patterns=(".git/*", ".env", ".env.*", "*.pem", "*.key"),
            max_read_lines=400,
            max_search_results=50,
            max_find_results=50,
        ),
        HarnessShellBridge(
            cwd=root,
            denied_commands=("rm", "mv", "cp", "curl", "wget", "git"),
            default_timeout=5.0,
            max_output_chars=8000,
            persist_cwd=False,
            allow_interactive=False,
        ),
    ]
    if include_code_mode:
        bridges.append(HarnessCodeModeBridge(max_retries=2))
    return tuple(bridges)


def _instructions(*, include_code_mode: bool) -> str:
    if include_code_mode:
        return _INSTRUCTIONS + _CODE_MODE_INSTRUCTIONS
    return _INSTRUCTIONS


def _harness_model(*, instructions: str = _INSTRUCTIONS) -> str | CodexResponsesModel:
    configured_model = os.getenv("ACP_HARNESS_MODEL", "").strip()
    if configured_model:
        return configured_model
    configured_codex_model = os.getenv("ACP_HARNESS_CODEX_MODEL", "").strip()
    if not configured_codex_model:
        return _DEFAULT_MODEL_NAME
    try:
        from codex_auth_helper import create_codex_responses_model
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(
            "Install `codex-auth-helper`, unset `ACP_HARNESS_CODEX_MODEL`, or set "
            "`ACP_HARNESS_MODEL` to a pydantic-ai model name before running the harness example.",
        ) from exc
    return create_codex_responses_model(
        configured_codex_model or _DEFAULT_CODEX_MODEL,
        instructions=instructions,
    )


def _build_config(*, include_code_mode: bool) -> AdapterConfig:
    return AdapterConfig(
        agent_name=_AGENT_NAME,
        agent_title="Harness Agent",
        session_store=FileSessionStore(_SESSION_STORE_ROOT),
        capability_bridges=_build_harness_bridges(
            _WORKSPACE_ROOT,
            include_code_mode=include_code_mode,
        ),
    )


def _build_agent(session: AcpSessionContext, *, include_code_mode: bool) -> Agent[None, str]:
    """Build a real model-backed agent wired with pydantic-ai-harness capabilities."""
    _ensure_workspace()
    harness_bridges = _build_harness_bridges(
        _WORKSPACE_ROOT,
        include_code_mode=include_code_mode,
    )
    bridge_contributions = AgentBridgeBuilder(
        session=session,
        capability_bridges=harness_bridges,
    ).build()
    return Agent(
        _harness_model(instructions=_instructions(include_code_mode=include_code_mode)),
        name=_AGENT_NAME,
        capabilities=bridge_contributions.capabilities,
        instructions=_instructions(include_code_mode=include_code_mode),
    )


def _build_agent_factory(
    *,
    include_code_mode: bool,
) -> AgentFactory[None, str]:
    def factory(session: AcpSessionContext) -> Agent[None, str]:
        return _build_agent(session, include_code_mode=include_code_mode)

    return factory


def agent_factory(session: AcpSessionContext) -> Agent[None, str]:
    return _build_agent(session, include_code_mode=False)


def _parse_args(argv: Sequence[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Run the ACP Kit harness example.")
    parser.add_argument(
        "--codemode",
        action="store_true",
        help="Enable the pydantic-ai-harness CodeMode capability bridge for this run.",
    )
    return parser.parse_args(argv)


_HARNESS_BRIDGES: Final[tuple[CapabilityBridge, ...]] = _build_harness_bridges(_WORKSPACE_ROOT)

config = _build_config(include_code_mode=False)
acp_agent = create_acp_agent(agent_factory=agent_factory, config=config)


def main(argv: Sequence[str] | None = None) -> None:
    args = _parse_args(()) if argv is None else _parse_args(argv)
    runtime_config = config
    runtime_agent_factory: AgentFactory[None, str] = agent_factory
    if args.codemode:
        runtime_config = _build_config(include_code_mode=True)
        runtime_agent_factory = _build_agent_factory(include_code_mode=True)
    _ensure_workspace()
    run_acp(agent_factory=runtime_agent_factory, config=runtime_config)


if __name__ == "__main__":  # pragma: no cover
    main(sys.argv[1:])



from __future__ import annotations as _annotations

from .adapter import LangChainAcpAgent
from .server import create_acp_agent, run_acp

__all__ = ("LangChainAcpAgent", "create_acp_agent", "run_acp")



from __future__ import annotations as _annotations

from dataclasses import dataclass
from typing import Final, Generic, Literal, TypeVar, cast

from acp.schema import SessionConfigOptionSelect, SessionConfigSelectOption, SessionMode
from pydantic_ai.capabilities import PrepareOutputTools, PrepareTools
from pydantic_ai.tools import RunContext, ToolDefinition, ToolsPrepareFunc

from .._slash_commands import validate_mode_command_ids
from ..agent_types import RuntimeAgent
from ..awaitables import resolve_value
from ..providers import ConfigOption, ModeState
from ..session.state import AcpSessionContext, JsonValue
from .base import BufferedCapabilityBridge

AgentDepsT = TypeVar("AgentDepsT", contravariant=True)
PlanGenerationType = Literal["tools", "structured"]

_PLAN_GENERATION_CONFIG_OPTIONS: Final[tuple[PlanGenerationType, ...]] = (
    "tools",
    "structured",
)

__all__ = (
    "PlanGenerationType",
    "PrepareOutputToolsBridge",
    "PrepareOutputToolsMode",
    "PrepareToolsBridge",
    "PrepareToolsMode",
)


@dataclass(slots=True, frozen=True, kw_only=True)
class PrepareToolsMode(Generic[AgentDepsT]):
    id: str
    name: str
    prepare_func: ToolsPrepareFunc[AgentDepsT]
    description: str | None = None
    plan_mode: bool = False
    plan_tools: bool = False


@dataclass(slots=True, frozen=True, kw_only=True)
class PrepareOutputToolsMode(Generic[AgentDepsT]):
    id: str
    name: str
    prepare_func: ToolsPrepareFunc[AgentDepsT]
    description: str | None = None


@dataclass(slots=True, kw_only=True)
class PrepareToolsBridge(BufferedCapabilityBridge, Generic[AgentDepsT]):
    metadata_key: str | None = "prepare_tools"
    default_mode_id: str
    modes: list[PrepareToolsMode[AgentDepsT]]
    mode_config_key: str = "mode"
    plan_generation_config_id: str = "plan_generation_type"
    plan_generation_config_name: str = "Plan Generation"
    plan_generation_config_description: str = "How plan mode records ACP plan state."
    default_plan_generation_type: PlanGenerationType = "structured"

    def __post_init__(self) -> None:
        if not self.modes:
            raise ValueError("PrepareToolsBridge requires at least one mode.")
        validate_mode_command_ids(mode.id for mode in self.modes)
        mode_ids = {mode.id for mode in self.modes}
        if self.default_mode_id not in mode_ids:
            raise ValueError("PrepareToolsBridge default mode must match one of the modes.")
        if sum(1 for mode in self.modes if mode.plan_mode) > 1:
            raise ValueError("PrepareToolsBridge supports at most one `plan_mode=True` mode.")
        if self.default_plan_generation_type not in _PLAN_GENERATION_CONFIG_OPTIONS:
            raise ValueError("PrepareToolsBridge default plan generation type is invalid.")

    def build_prepare_tools(self, session: AcpSessionContext) -> ToolsPrepareFunc[AgentDepsT]:
        async def prepare_tools(
            ctx: RunContext[AgentDepsT],
            tool_defs: list[ToolDefinition],
        ) -> list[ToolDefinition]:
            mode = self._require_mode(self._current_mode_id(session))
            try:
                prepared = mode.prepare_func(ctx, list(tool_defs))
                resolved = await resolve_value(prepared)
            except Exception as error:
                self._record_failed_event(
                    session,
                    title=f"prepare_tools.{mode.id}",
                    raw_output=str(error),
                )
                raise

            next_tool_defs = list(tool_defs if resolved is None else resolved)
            self._record_completed_event(
                session,
                title=f"prepare_tools.{mode.id}",
                raw_output=f"tools={len(next_tool_defs)}/{len(tool_defs)}",
            )
            return next_tool_defs

        return prepare_tools

    def build_capability(self, session: AcpSessionContext) -> PrepareTools[AgentDepsT]:
        return PrepareTools(self.build_prepare_tools(session))

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[PrepareTools[AgentDepsT], ...]:
        return (self.build_capability(session),)

    def get_mode_state(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> ModeState:
        del agent
        return ModeState(
            modes=[
                SessionMode(
                    id=mode.id,
                    name=mode.name,
                    description=mode.description,
                )
                for mode in self.modes
            ],
            current_mode_id=self._current_mode_id(session),
        )

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del agent
        modes: list[JsonValue] = [
            {
                "description": mode.description,
                "id": mode.id,
                "name": mode.name,
                "plan_mode": mode.plan_mode,
                "plan_tools": mode.plan_tools,
            }
            for mode in self.modes
        ]
        metadata: dict[str, JsonValue] = {
            "current_mode_id": self._current_mode_id(session),
            "modes": modes,
        }
        if self.supports_plan_generation_selection():
            metadata["current_plan_generation_type"] = self.current_plan_generation_type(session)
            metadata["supported_plan_generation_types"] = list(_PLAN_GENERATION_CONFIG_OPTIONS)
        return metadata

    def get_config_options(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> list[ConfigOption]:
        del agent
        if not self.supports_plan_generation_selection():
            return []
        return [
            SessionConfigOptionSelect(
                id=self.plan_generation_config_id,
                name=self.plan_generation_config_name,
                category="agent",
                description=self.plan_generation_config_description,
                type="select",
                current_value=self.current_plan_generation_type(session),
                options=[
                    SessionConfigSelectOption(
                        value=value,
                        name="Tool Plans" if value == "tools" else "Structured Plans",
                    )
                    for value in _PLAN_GENERATION_CONFIG_OPTIONS
                ],
            ),
        ]

    def set_mode(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
        mode_id: str,
    ) -> ModeState | None:
        del agent
        if self._find_mode(mode_id) is None:
            return None
        session.config_values[self.mode_config_key] = mode_id
        return ModeState(
            modes=[
                SessionMode(
                    id=mode.id,
                    name=mode.name,
                    description=mode.description,
                )
                for mode in self.modes
            ],
            current_mode_id=mode_id,
        )

    def set_config_option(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
        config_id: str,
        value: str | bool,
    ) -> list[ConfigOption] | None:
        if (
            not self.supports_plan_generation_selection()
            or config_id != self.plan_generation_config_id
            or not isinstance(value, str)
            or value not in _PLAN_GENERATION_CONFIG_OPTIONS
        ):
            return None
        if value == self.default_plan_generation_type:
            session.config_values.pop(self.plan_generation_config_id, None)
        else:
            session.config_values[self.plan_generation_config_id] = value
        return self.get_config_options(session, agent)

    def _current_mode_id(self, session: AcpSessionContext) -> str:
        configured_mode = session.config_values.get(self.mode_config_key)
        if isinstance(configured_mode, str) and self._find_mode(configured_mode) is not None:
            return configured_mode
        return self.default_mode_id

    def _find_mode(self, mode_id: str) -> PrepareToolsMode[AgentDepsT] | None:
        for mode in self.modes:
            if mode.id == mode_id:
                return mode
        return None

    def _require_mode(self, mode_id: str) -> PrepareToolsMode[AgentDepsT]:
        mode = self._find_mode(mode_id)
        if mode is None:
            raise ValueError(f"Unknown prepare-tools mode: {mode_id!r}")
        return mode

    def current_mode(self, session: AcpSessionContext) -> PrepareToolsMode[AgentDepsT]:
        return self._require_mode(self._current_mode_id(session))

    def supports_plan_generation_selection(self) -> bool:
        return any(mode.plan_mode for mode in self.modes)

    def current_plan_generation_type(self, session: AcpSessionContext) -> PlanGenerationType:
        configured_value = session.config_values.get(self.plan_generation_config_id)
        if (
            isinstance(configured_value, str)
            and configured_value in _PLAN_GENERATION_CONFIG_OPTIONS
        ):
            return cast("PlanGenerationType", configured_value)
        return self.default_plan_generation_type

    def uses_tool_plan_generation(self, session: AcpSessionContext) -> bool:
        return self.current_plan_generation_type(session) == "tools"

    def uses_structured_plan_generation(self, session: AcpSessionContext) -> bool:
        return self.current_plan_generation_type(session) == "structured"

    def is_plan_mode(self, session: AcpSessionContext) -> bool:
        return self.current_mode(session).plan_mode

    def supports_plan_tools(self, session: AcpSessionContext) -> bool:
        mode = self.current_mode(session)
        return mode.plan_mode or mode.plan_tools

    def supports_plan_write_tools(self, session: AcpSessionContext) -> bool:
        mode = self.current_mode(session)
        return mode.plan_tools or (mode.plan_mode and self.uses_tool_plan_generation(session))

    def supports_plan_progress(self, session: AcpSessionContext) -> bool:
        return self.current_mode(session).plan_tools


@dataclass(slots=True, kw_only=True)
class PrepareOutputToolsBridge(BufferedCapabilityBridge, Generic[AgentDepsT]):
    metadata_key: str | None = "prepare_output_tools"
    default_mode_id: str
    modes: list[PrepareOutputToolsMode[AgentDepsT]]
    mode_config_key: str = "prepare_output_tools_mode"

    def __post_init__(self) -> None:
        if not self.modes:
            raise ValueError("PrepareOutputToolsBridge requires at least one mode.")
        validate_mode_command_ids(mode.id for mode in self.modes)
        mode_ids = {mode.id for mode in self.modes}
        if self.default_mode_id not in mode_ids:
            raise ValueError("PrepareOutputToolsBridge default mode must match one of the modes.")

    def build_prepare_output_tools(
        self,
        session: AcpSessionContext,
    ) -> ToolsPrepareFunc[AgentDepsT]:
        async def prepare_output_tools(
            ctx: RunContext[AgentDepsT],
            tool_defs: list[ToolDefinition],
        ) -> list[ToolDefinition]:
            mode = self._require_mode(self._current_mode_id(session))
            try:
                prepared = mode.prepare_func(ctx, list(tool_defs))
                resolved = await resolve_value(prepared)
            except Exception as error:
                self._record_failed_event(
                    session,
                    title=f"prepare_output_tools.{mode.id}",
                    raw_output=str(error),
                )
                raise

            next_tool_defs = list(tool_defs if resolved is None else resolved)
            self._record_completed_event(
                session,
                title=f"prepare_output_tools.{mode.id}",
                raw_output=f"tools={len(next_tool_defs)}/{len(tool_defs)}",
            )
            return next_tool_defs

        return prepare_output_tools

    def build_capability(self, session: AcpSessionContext) -> PrepareOutputTools[AgentDepsT]:
        return PrepareOutputTools(self.build_prepare_output_tools(session))

    def build_agent_capabilities(
        self,
        session: AcpSessionContext,
    ) -> tuple[PrepareOutputTools[AgentDepsT], ...]:
        return (self.build_capability(session),)

    def get_mode_state(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> ModeState:
        del agent
        return ModeState(
            modes=[
                SessionMode(
                    id=mode.id,
                    name=mode.name,
                    description=mode.description,
                )
                for mode in self.modes
            ],
            current_mode_id=self._current_mode_id(session),
        )

    def get_session_metadata(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
    ) -> dict[str, JsonValue]:
        del agent
        modes: list[JsonValue] = [
            {
                "description": mode.description,
                "id": mode.id,
                "name": mode.name,
            }
            for mode in self.modes
        ]
        return {
            "current_mode_id": self._current_mode_id(session),
            "modes": modes,
        }

    def set_mode(
        self,
        session: AcpSessionContext,
        agent: RuntimeAgent,
        mode_id: str,
    ) -> ModeState | None:
        del agent
        if self._find_mode(mode_id) is None:
            return None
        session.config_values[self.mode_config_key] = mode_id
        return ModeState(
            modes=[
                SessionMode(
                    id=mode.id,
                    name=mode.name,
                    description=mode.description,
                )
                for mode in self.modes
            ],
            current_mode_id=mode_id,
        )

    def _current_mode_id(self, session: AcpSessionContext) -> str:
        configured_mode = session.config_values.get(self.mode_config_key)
        if isinstance(configured_mode, str) and self._find_mode(configured_mode) is not None:
            return configured_mode
        return self.default_mode_id

    def _find_mode(self, mode_id: str) -> PrepareOutputToolsMode[AgentDepsT] | None:
        for mode in self.modes:
            if mode.id == mode_id:
                return mode
        return None

    def _require_mode(self, mode_id: str) -> PrepareOutputToolsMode[AgentDepsT]:
        mode = self._find_mode(mode_id)
        if mode is None:
            raise ValueError(f"Unknown prepare output tools mode: {mode_id!r}")
        return mode



from __future__ import annotations as _annotations

import re
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any

from acp.schema import (
    AvailableCommand,
    AvailableCommandInput,
    SessionMode,
    SessionModelState,
    SessionModeState,
    UnstructuredCommandInput,
)

from .._slash_commands import (
    MCP_SERVERS_COMMAND_NAME,
    MODEL_COMMAND_NAME,
    RESERVED_SLASH_COMMAND_NAMES,
    TOOLS_COMMAND_NAME,
    validate_mode_command_ids,
)
from ..session.state import AcpSessionContext, JsonValue

__all__ = (
    "McpServerInfo",
    "SlashCommand",
    "ToolInfo",
    "build_available_commands",
    "extract_session_mcp_servers",
    "list_graph_tools",
    "parse_slash_command",
    "render_mcp_server_listing",
    "render_mode_message",
    "render_model_message",
    "render_tool_listing",
    "validate_custom_commands",
)

_COMMAND_NAME_PATTERN = re.compile(r"^[a-z][a-z0-9-]*$")


@dataclass(slots=True, frozen=True, kw_only=True)
class SlashCommand:
    name: str
    argument: str | None = None


@dataclass(slots=True, frozen=True, kw_only=True)
class ToolInfo:
    name: str
    description: str | None


@dataclass(slots=True, frozen=True, kw_only=True)
class McpServerInfo:
    name: str
    transport: str
    target: str
    source: str


def build_available_commands(
    *,
    mode_state: SessionModeState | None,
    model_state: SessionModelState | None,
    custom_commands: Sequence[AvailableCommand] | None = None,
) -> list[AvailableCommand]:
    commands: list[AvailableCommand] = []
    if mode_state is not None:
        validate_mode_command_ids(mode.id for mode in mode_state.available_modes)
        commands.extend(_mode_commands(mode_state.available_modes))
    if model_state is not None:
        commands.append(
            AvailableCommand(
                name=MODEL_COMMAND_NAME,
                description="Show the current session model, or set it with a provider:model value.",
                input=AvailableCommandInput(root=UnstructuredCommandInput(hint="provider:model")),
            ),
        )
    commands.extend(
        [
            AvailableCommand(
                name=TOOLS_COMMAND_NAME,
                description="List the tools currently registered on the active graph.",
            ),
            AvailableCommand(
                name=MCP_SERVERS_COMMAND_NAME,
                description="List MCP servers attached to the current session.",
            ),
        ],
    )
    if custom_commands:
        validate_custom_commands(custom_commands, mode_state=mode_state)
        commands.extend(custom_commands)
    return commands


def validate_custom_commands(
    commands: Sequence[AvailableCommand],
    *,
    mode_state: SessionModeState | None,
) -> None:
    normalized_names: list[str] = []
    for command in commands:
        normalized_name = command.name.strip().lower()
        if command.name != normalized_name:
            raise ValueError(
                f"Slash command name {command.name!r} must already be normalized as "
                "a lowercase slash command id.",
            )
        if not _COMMAND_NAME_PATTERN.fullmatch(normalized_name):
            raise ValueError(f"Slash command name {command.name!r} must match ^[a-z][a-z0-9-]*$.")
        normalized_names.append(normalized_name)
    duplicate_names = sorted(
        name for name in set(normalized_names) if normalized_names.count(name) > 1
    )
    if duplicate_names:
        raise ValueError(
            "Custom slash command names must be unique after normalization. "
            f"Duplicate ids: {', '.join(duplicate_names)}.",
        )
    reserved_names = sorted(set(normalized_names) & RESERVED_SLASH_COMMAND_NAMES)
    if reserved_names:
        raise ValueError(
            "Custom slash command names cannot reuse reserved slash command names "
            f"({', '.join(reserved_names)}).",
        )
    if mode_state is None:
        return
    mode_ids = {mode.id.strip().lower() for mode in mode_state.available_modes}
    conflicting_mode_ids = sorted(set(normalized_names) & mode_ids)
    if conflicting_mode_ids:
        raise ValueError(
            "Custom slash command names cannot reuse active mode ids "
            f"({', '.join(conflicting_mode_ids)}).",
        )


def parse_slash_command(prompt_text: str) -> SlashCommand | None:
    stripped = prompt_text.strip()
    if not stripped.startswith("/"):
        return None
    command_text = stripped[1:]
    if not command_text.strip():
        return None
    name, _, remainder = command_text.partition(" ")
    normalized_name = name.strip().lower()
    if not normalized_name:
        return None
    argument = remainder.strip() or None
    return SlashCommand(name=normalized_name, argument=argument)


def render_mode_message(current_mode_id: str | None) -> str:
    if current_mode_id is None:
        return "Current mode: unavailable"
    return f"Current mode: {current_mode_id}"


def render_model_message(current_model_id: str | None) -> str:
    if current_model_id is None:
        return "Current model: unavailable"
    return f"Current model: {current_model_id}"


def render_tool_listing(tool_infos: list[ToolInfo]) -> str:
    if not tool_infos:
        return "No tools are currently registered."
    lines = ["Available tools:"]
    for tool_info in tool_infos:
        if tool_info.description is not None:
            lines.append(f"- {tool_info.name}: {tool_info.description}")
        else:
            lines.append(f"- {tool_info.name}")
    return "\n".join(lines)


def render_mcp_server_listing(server_infos: list[McpServerInfo]) -> str:
    if not server_infos:
        return "No MCP servers are currently attached."
    lines = ["MCP servers:"]
    for server_info in server_infos:
        lines.append(
            f"- {server_info.name} ({server_info.transport}, {server_info.source}): "
            f"{server_info.target}",
        )
    return "\n".join(lines)


def list_graph_tools(graph: Any) -> list[ToolInfo]:
    graph_view_factory = getattr(graph, "get_graph", None)
    if not callable(graph_view_factory):
        return []
    graph_view = graph_view_factory()
    nodes = getattr(graph_view, "nodes", None)
    if not isinstance(nodes, dict):
        return []
    tool_infos: list[ToolInfo] = []
    for node in nodes.values():
        tool_node = getattr(node, "data", None)
        tools_by_name = getattr(tool_node, "_tools_by_name", None)
        if not isinstance(tools_by_name, dict):
            continue
        string_items = [
            (name, tool) for name, tool in tools_by_name.items() if isinstance(name, str)
        ]
        for name, tool in sorted(string_items, key=lambda item: item[0]):
            description = getattr(tool, "description", None)
            tool_infos.append(
                ToolInfo(
                    name=name,
                    description=description if isinstance(description, str) else None,
                ),
            )
    return tool_infos


def extract_session_mcp_servers(session: AcpSessionContext) -> list[McpServerInfo]:
    server_infos: list[McpServerInfo] = []
    seen: set[tuple[str, str, str]] = set()
    for raw_server in session.mcp_servers:
        server_info = _mcp_server_info_from_session_payload(raw_server)
        if server_info is None:
            continue
        dedupe_key = (
            server_info.name,
            server_info.transport,
            server_info.target,
        )
        if dedupe_key in seen:
            continue
        seen.add(dedupe_key)
        server_infos.append(server_info)
    metadata_servers = session.metadata.get("mcp")
    if not isinstance(metadata_servers, dict):
        return server_infos
    raw_servers = metadata_servers.get("servers")
    if not isinstance(raw_servers, list):
        return server_infos
    for raw_server in raw_servers:
        server_info = _mcp_server_info_from_bridge_metadata(raw_server)
        if server_info is None:
            continue
        dedupe_key = (
            server_info.name,
            server_info.transport,
            server_info.target,
        )
        if dedupe_key in seen:
            continue
        seen.add(dedupe_key)
        server_infos.append(server_info)
    return server_infos


def _mode_commands(modes: Sequence[SessionMode]) -> list[AvailableCommand]:
    return [
        AvailableCommand(
            name=mode.id,
            description=mode.description or f"Switch the active session into {mode.name} mode.",
        )
        for mode in modes
    ]


def _mcp_server_info_from_session_payload(
    raw_server: dict[str, JsonValue],
) -> McpServerInfo | None:
    name = raw_server.get("name")
    if not isinstance(name, str) or not name:
        return None
    transport = raw_server.get("type")
    if not isinstance(transport, str) or not transport:
        transport = raw_server.get("transport")
    if not isinstance(transport, str) or not transport:
        return None
    if transport == "stdio":
        command = raw_server.get("command")
        args = raw_server.get("args")
        rendered_args = (
            " ".join(item for item in args if isinstance(item, str))
            if isinstance(args, list)
            else ""
        )
        target = command if isinstance(command, str) else "<stdio>"
        if rendered_args:
            target = f"{target} {rendered_args}"
    else:
        url = raw_server.get("url")
        target = url if isinstance(url, str) and url else f"<{transport}>"
    return McpServerInfo(
        name=name,
        transport=transport,
        target=target,
        source="session",
    )


def _mcp_server_info_from_bridge_metadata(
    raw_server: JsonValue,
) -> McpServerInfo | None:
    raw_server_dict = _string_key_dict(raw_server)
    if raw_server_dict is None:
        return None
    name = raw_server_dict.get("name")
    transport = raw_server_dict.get("transport")
    if not isinstance(name, str) or not isinstance(transport, str):
        return None
    url = raw_server_dict.get("url")
    description = raw_server_dict.get("description")
    target_parts = [value for value in (url, description) if isinstance(value, str) and value]
    target = " | ".join(target_parts) if target_parts else f"<{transport}>"
    return McpServerInfo(
        name=name,
        transport=transport,
        target=target,
        source="bridge",
    )


def _string_key_dict(value: JsonValue) -> dict[str, JsonValue] | None:
    if not isinstance(value, dict):
        return None
    string_key_items = [(key, item) for key, item in value.items() if isinstance(key, str)]
    if len(string_key_items) != len(value):
        return None
    return dict(string_key_items)



from __future__ import annotations as _annotations

from collections.abc import Sequence
from dataclasses import dataclass, field

from acp.schema import ModelInfo, SessionMode

from ._version import __version__
from .approvals import ApprovalBridge, NativeApprovalBridge
from .bridges import CapabilityBridge
from .event_projection import EventProjectionMap
from .plan import PlanGenerationType
from .projection import DefaultToolClassifier, ProjectionMap, ToolClassifier
from .prompt_capabilities import AdapterPromptCapabilities
from .providers import (
    ConfigOptionsProvider,
    NativePlanPersistenceProvider,
    PlanProvider,
    SessionModelsProvider,
    SessionModesProvider,
)
from .serialization import DefaultOutputSerializer, OutputSerializer
from .session.store import MemorySessionStore, SessionStore
from .slash import SlashCommandProvider

DEFAULT_AGENT_NAME = "langchain-acp"
DEFAULT_AGENT_TITLE = "LangChain ACP"
DEFAULT_AGENT_VERSION = __version__

__all__ = (
    "DEFAULT_AGENT_NAME",
    "DEFAULT_AGENT_TITLE",
    "DEFAULT_AGENT_VERSION",
    "AdapterConfig",
)


@dataclass(slots=True, kw_only=True)
class AdapterConfig:
    agent_name: str = DEFAULT_AGENT_NAME
    agent_title: str = DEFAULT_AGENT_TITLE
    agent_version: str = DEFAULT_AGENT_VERSION
    approval_bridge: ApprovalBridge | None = field(default_factory=NativeApprovalBridge)
    available_models: list[ModelInfo] = field(default_factory=list)
    available_modes: list[SessionMode] = field(default_factory=list)
    capability_bridges: Sequence[CapabilityBridge] = field(default_factory=tuple)
    config_options_provider: ConfigOptionsProvider | None = None
    default_model_id: str | None = None
    default_mode_id: str | None = None
    default_plan_generation_type: PlanGenerationType = "structured"
    enable_plan_progress_tools: bool = False
    event_projection_maps: Sequence[EventProjectionMap] = field(default_factory=tuple)
    models_provider: SessionModelsProvider | None = None
    modes_provider: SessionModesProvider | None = None
    native_plan_additional_instructions: str | None = None
    native_plan_persistence_provider: NativePlanPersistenceProvider | None = None
    output_serializer: OutputSerializer = field(default_factory=DefaultOutputSerializer)
    plan_mode_id: str | None = None
    plan_provider: PlanProvider | None = None
    projection_maps: Sequence[ProjectionMap] = field(default_factory=tuple)
    prompt_capabilities: AdapterPromptCapabilities = field(
        default_factory=AdapterPromptCapabilities,
    )
    replay_history_on_load: bool = True
    slash_command_provider: SlashCommandProvider | None = None
    session_store: SessionStore = field(default_factory=MemorySessionStore)
    tool_classifier: ToolClassifier = field(default_factory=DefaultToolClassifier)
